#!/usr/bin/env pmpython
#
# Copyright (c) 2023 Oracle and/or its affiliates.
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# pylint: disable=bad-whitespace,too-many-lines,bad-continuation
# pylint: disable=too-many-arguments,too-many-positional-arguments
# pylint: disable=redefined-outer-name,unnecessary-lambda
#
import signal
import sys
import time
from pcp import pmapi, pmcc
from cpmapi import PM_CONTEXT_ARCHIVE

SYS_MECTRICS= ["kernel.uname.sysname","kernel.uname.release",
               "kernel.uname.nodename","kernel.uname.machine","hinv.ncpu"]
SLABSTAT_METRICS = ["mem.slabinfo.slabs.active",
                    "mem.slabinfo.slabs.pages_per_slab","mem.slabinfo.slabs.total_size",
                    "mem.slabinfo.slabs.objects_per_slab","mem.slabinfo.objects.active",
                    "mem.slabinfo.objects.total","mem.slabinfo.objects.size",
                    "mem.slabinfo.slabs.total" ]
ALL_METRICS = SLABSTAT_METRICS + SYS_MECTRICS
def adjust_length(name):
    return name.ljust(25)
class ReportingMetricRepository:

    def __init__(self,group):
        self.group=group
        self.current_cached_values = {}

    def __sorted(self,data):
        return dict(sorted(data.items(), key=lambda item: item[0].lower()))

    def __fetch_current_value(self,metric):
        val=dict(map(lambda x: (x[1], x[2]), self.group[metric].netValues))
        val=self.__sorted(val)
        return dict(val)

    def current_value(self,metric):
        if not metric in self.group:
            return None
        if self.current_cached_values.get(metric) is None:
            first_value=self.__fetch_current_value(metric)
            self.current_cached_values[metric]=first_value
        return self.current_cached_values[metric]

class SlabStatUtil:
    def __init__(self,metrics_repository):
        self.__metric_repository=metrics_repository
        self.report=ReportingMetricRepository(self.__metric_repository)

    def active_objs(self):
        return self.report.current_value('mem.slabinfo.objects.active')

    def objs_size(self):
        return self.report.current_value('mem.slabinfo.objects.size')

    def obj_total(self):
        return self.report.current_value('mem.slabinfo.objects.total')

    def slab_total_size(self):
        return self.report.current_value('mem.slabinfo.slabs.total_size')

    def objects_per_slab(self):
        return self.report.current_value('mem.slabinfo.slabs.objects_per_slab')

    def page_per_slab(self):
        return self.report.current_value('mem.slabinfo.slabs.pages_per_slab')

    def slab_active(self):
        return self.report.current_value('mem.slabinfo.slabs.active')

    def slabs_total(self):
        return self.report.current_value('mem.slabinfo.slabs.total')

    def names(self):
        data = self.report.current_value('mem.slabinfo.objects.active')
        return data.keys()

class SlabinfoReport(pmcc.MetricGroupPrinter):
    def __init__(self,opts,group):
        self.opts = opts
        self.group = group
        self.samples = opts.samples
        self.context = opts.context

    def __get_ncpu(self, group):
        return group['hinv.ncpu'].netValues[0][2]

    def __print_machine_info(self, context):
        timestamp = self.group.pmLocaltime(context.timestamp.tv_sec)
        # Please check strftime(3) for different formatting options.
        # Also check TZ and LC_TIME environment variables for more
        # information on how to override the default formatting of
        # the date display in the header
        time_string = time.strftime("%x", timestamp.struct_time())
        header_string = ''
        header_string += context['kernel.uname.sysname'].netValues[0][2] + '  '
        header_string += context['kernel.uname.release'].netValues[0][2] + '  '
        header_string += '(' + context['kernel.uname.nodename'].netValues[0][2] + ')  '
        header_string += time_string + '  '
        header_string += context['kernel.uname.machine'].netValues[0][2] + '  '
        print("%s  (%s CPU)" % (header_string, self.__get_ncpu(context)))

    def __print_header(self,header_indentation,value_indentation):
        print("TimeStamp"+ " "+header_indentation + "Name" + " "*22 +value_indentation+ "active_objs" +
              value_indentation +"num_objs" + value_indentation + "objsize byte" + value_indentation +
              "objperslab" + value_indentation +"pagesperslab" + value_indentation + "active_slabs" +
              value_indentation + "num_slabs")

    def __print_values(self,timestamp,header_indentation,\
                     value_indentation,slabstatus):
        names=slabstatus.names()
        active_objects=slabstatus.active_objs()
        num_objs=slabstatus.objs_size()
        objsize=slabstatus.obj_total()
        objperslab=slabstatus.objects_per_slab()
        pageperslab=slabstatus.page_per_slab()
        active_slabs=slabstatus.slab_active()
        num_slabs=slabstatus.slabs_total()
        for name in names:
            data = adjust_length(name) if len(name) < 25 else name
            print("%s %s %s %s %8d %s %8d %s %8d %s %7d %s %9d %s %8d %s %8d"
                                    % (timestamp, header_indentation,data, value_indentation, active_objects[name],
                                       value_indentation,objsize[name],value_indentation, num_objs[name],
                                       value_indentation, objperslab[name],value_indentation,pageperslab[name],
                                       value_indentation, active_slabs[name], value_indentation, num_slabs[name]))

    def print_report(self,group,timestamp,header_indentation,value_indentation,manager_slabinfo):
        def __print_slab_status():
            slabstatus = SlabStatUtil(manager_slabinfo)
            if slabstatus.names():
                try:
                    self.__print_machine_info(group)
                    self.__print_header(header_indentation, value_indentation)
                    self.__print_values(timestamp, header_indentation, value_indentation, slabstatus)
                except IndexError:
                    print("Incorrect machine info due to some missing metrics")
                return
            else:
                pass

        if self.context != PM_CONTEXT_ARCHIVE and self.samples is None:
            __print_slab_status()
            sys.exit(0)
        elif self.context == PM_CONTEXT_ARCHIVE and self.samples is None:
            __print_slab_status()
        elif self.samples >=1:
            __print_slab_status()
            self.samples-=1
        else:
            pass

    def report(self, manager):
        group = manager["sysinfo"]
        self.samples = self.opts.pmGetOptionSamples()
        t_s = group.contextCache.pmLocaltime(int(group.timestamp))
        timestamp = time.strftime(SlabinfoOptions.timefmt, t_s.struct_time())
        header_indentation = "        " if len(timestamp) < 9 else (len(timestamp) - 7) * " "
        value_indentation = ((len(header_indentation) + 9) - len(timestamp)) * " "
        self.print_report(group,timestamp,header_indentation,value_indentation,manager['slabinfo'])

class SlabinfoOptions(pmapi.pmOptions):
    timefmt = "%H:%M:%S"
    def __init__(self):
        pmapi.pmOptions.__init__(self, "a:s:Z:zV?")
        self.pmSetLongOptionHeader("General options")
        self.pmSetLongOptionHostZone()
        self.pmSetLongOptionTimeZone()
        self.pmSetLongOptionHelp()
        self.pmSetLongOptionSamples()
        self.pmSetLongOptionVersion()
        self.samples=None
        self.context=None

if __name__ == '__main__':
    try:
        opts = SlabinfoOptions()
        mngr = pmcc.MetricGroupManager.builder(opts,sys.argv)
        opts.context=mngr.type
        missing = mngr.checkMissingMetrics(ALL_METRICS)
        if missing is not None:
            sys.stderr.write('Error: not all required metrics are available\nMissing %s\n' % missing)
            sys.exit(1)
        mngr["slabinfo"] = SLABSTAT_METRICS
        mngr["sysinfo"] = SYS_MECTRICS
        mngr["allinfo"]=ALL_METRICS
        mngr.printer = SlabinfoReport(opts,mngr)
        sts = mngr.run()
        sys.exit(sts)
    except pmapi.pmErr as error:
        sys.stderr.write('%s\n' % (error.message()))
    except pmapi.pmUsageErr as usage:
        usage.message()
        sys.exit(1)
    except IOError:
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    except KeyboardInterrupt:
        pass



#!/usr/bin/pmpython
#
# Copyright (C) 2025 Red Hat.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# Create a PCP archive from vLLM benchmark results JSON file.
#

import argparse
import calendar
import json
import sys

from datetime import datetime, timedelta, timezone
from pcp import pmapi, pmi
from cpmapi import (PM_COUNT_ONE, PM_LABEL_CONTEXT,
                    PM_TIME_SEC, PM_TIME_MSEC,
                    PM_TYPE_DOUBLE, PM_TYPE_U64, PM_SEM_DISCRETE,
                    PM_TEXT_PMID, PM_TEXT_ONELINE, PM_TEXT_HELP)

def register_metrics(pcp) -> pmapi.pmInDom:
    domain = 510 # PCP domain number (vllm import)
    indom = pcp.pmiInDom(domain, 1) # benchmark iterations
    cluster = 0 # incremented for each metric group added
    item = 0 # incremented for each metric within a group

    nounit = pcp.pmiUnits(0, 0, 0, 0, 0, 0)
    unitone = pcp.pmiUnits(0, 0, 1, 0, 0, PM_COUNT_ONE)
    secunits = pcp.pmiUnits(0, 1, 0, 0, PM_TIME_SEC, 0)
    msecunits = pcp.pmiUnits(0, 1, 0, 0, PM_TIME_MSEC, 0)
    psecunits = pcp.pmiUnits(0, 1, 1, 0, PM_TIME_SEC, PM_COUNT_ONE)

    def add_metric(name, datatype, units, oneline=None, text=None) -> None:
        nonlocal item
        pmid = pcp.pmiID(domain, cluster, item)
        pcp.pmiAddMetric(name, pmid, datatype,
                     indom, PM_SEM_DISCRETE, units)
        if oneline:
            pcp.pmiPutText(PM_TEXT_PMID, PM_TEXT_ONELINE, pmid, oneline)
        if text:
            pcp.pmiPutText(PM_TEXT_PMID, PM_TEXT_HELP, pmid, text)
        item += 1

    pmns = 'vllm.bench.'
    add_metric(pmns + 'duration', PM_TYPE_DOUBLE, secunits,
        'Elapsed time in seconds for each vllm bench run completed')
    add_metric(pmns + 'request_rate', PM_TYPE_DOUBLE, psecunits,
        'Request processing rate goal, a benchmark input parameter',
        "Number of requests per second. If this is inf, then all requests "
        "are sent at time 0.  Otherwise vllm uses Poisson process or gamma "
        "distribution to synthesize the request arrival times.")
    add_metric(pmns + 'burstiness', PM_TYPE_DOUBLE, nounit,
        'Burstiness factor of request generation, a benchmark input parameter',
        "Burstiness only takes effect when request_rate is set. "
        "Default value is 1, which follows Poisson process. "
        "Otherwise, the request intervals follow a gamma distribution. "
        "A lower burstiness value (0 < burstiness < 1) results in more "
        "bursty requests. A higher burstiness value (burstiness > 1) "
        "results in a more uniform arrival of requests.")
    add_metric(pmns + 'max_concurrency', PM_TYPE_DOUBLE, unitone,
        'Maximum number of concurrent requests, a benchmark input parameter',
        "Can be used to help simulate an environment where a higher level "
        "component is enforcing a maximum number of concurrent requests.  "
        "While the --request-rate argument controls the rate requests are "
        "initiated, this argument will control how many are actually allowed "
        "to execute at a time.  This means that when used in combination, the "
        "actual request rate may be lower than specified with --request-rate, "
        "if the server is not processing requests fast enough to keep up.")
    add_metric(pmns + 'completed', PM_TYPE_U64, unitone,
               'Count of requests processed (requests-per-second as a rate)')
    add_metric(pmns + 'num_prompts', PM_TYPE_U64, unitone,
               'Count of all requests started irrespective of how they finish')

    add_metric(pmns + 'total_input_tokens', PM_TYPE_U64, unitone,
        'Total number of input tokens')
    add_metric(pmns + 'total_output_tokens', PM_TYPE_U64, unitone,
        'Total number of output tokens')
    add_metric(pmns + 'request_throughput', PM_TYPE_DOUBLE, psecunits,
        'Total number of requests handled per second')
    add_metric(pmns + 'request_goodput', PM_TYPE_DOUBLE, psecunits,
        'Total number of requests meeting service level objectives per second',
        "")
    add_metric(pmns + 'output_throughput', PM_TYPE_DOUBLE, psecunits,
        'Total number of output tokens per second')
    add_metric(pmns + 'total_token_throughput', PM_TYPE_DOUBLE, psecunits,
        'Total number of tokens per second')
    add_metric(pmns + 'max_output_tokens_per_s', PM_TYPE_DOUBLE, secunits,
        'Maximum rate of output tokens')
    add_metric(pmns + 'max_concurrent_requests', PM_TYPE_U64, nounit,
        'Maximum number of concurrent requests')
    add_metric(pmns + 'mean_ttft_ms', PM_TYPE_DOUBLE, msecunits,
        'Average time taken to generate the first token of the output',
        'Indicates initial response time of the model, crucial for user-\n' +
        'facing applications.')
    add_metric(pmns + 'median_ttft_ms', PM_TYPE_DOUBLE, msecunits,
        'Median time taken to generate the first token of the output',
        'Indicates initial response time of the model, crucial for user-\n' +
        'facing applications.')
    add_metric(pmns + 'std_ttft_ms', PM_TYPE_DOUBLE, msecunits,
        'Standard deviation of time taken to generate first output token')
    add_metric(pmns + 'p99_ttft_ms', PM_TYPE_DOUBLE, msecunits,
        '99th percentile time taken to generate first output token')
    add_metric(pmns + 'mean_tpot_ms', PM_TYPE_DOUBLE, msecunits,
        'Average time taken to generate each output token',
        'Provides a detailed view of the models token generation efficiency.')
    add_metric(pmns + 'median_tpot_ms', PM_TYPE_DOUBLE, msecunits,
        'Median time taken to generate each output token',
        'Provides a detailed view of the models token generation efficiency.')
    add_metric(pmns + 'std_tpot_ms', PM_TYPE_DOUBLE, msecunits,
        'Standard deviation of time taken to generate each output token')
    add_metric(pmns + 'p99_tpot_ms', PM_TYPE_DOUBLE, msecunits,
        '99th percentile of time taken to generate each output token')
    add_metric(pmns + 'mean_itl_ms', PM_TYPE_DOUBLE, msecunits,
        'Average time between generating consecutive tokens in the output',
        'Helps assess the smoothness and speed of token generation.')
    add_metric(pmns + 'median_itl_ms', PM_TYPE_DOUBLE, msecunits,
        'Median time between generating consecutive tokens in the output',
        'Helps assess the smoothness and speed of token generation.')
    add_metric(pmns + 'std_itl_ms', PM_TYPE_DOUBLE, msecunits,
        'Standard deviation of time between consecutive output tokens')
    add_metric(pmns + 'p99_itl_ms', PM_TYPE_DOUBLE, msecunits,
        '99th percentile of time between consecutive output tokens')
    add_metric(pmns + 'mean_e2el_ms', PM_TYPE_DOUBLE, msecunits,
        'Average end-to-end request latency per request',
        "End-to-end latency is the time taken on the client side from "
        "sending a request to receiving the complete response.  This "
        "measure helps to assess the end-user perceived performance.")
    add_metric(pmns + 'median_e2el_ms', PM_TYPE_DOUBLE, msecunits,
        'Median end-to-end request latency per request',
        "End-to-end latency is the time taken on the client side from "
        "sending a request to receiving the complete response.  This "
        "measure helps to assess the end-user perceived performance.")
    add_metric(pmns + 'std_e2el_ms', PM_TYPE_DOUBLE, msecunits,
        'Standard deviation of end-to-end request latency')
    add_metric(pmns + 'p99_e2el_ms', PM_TYPE_DOUBLE, msecunits,
        '99th percentile of end-to-end request latency')

    return indom

def put_metric_values(pcp, values, instance) -> None:
    """ Scan through JSON data and extract values for each PCP metric;
        pmiPutValue inserts one value for one metric:inst at one time.
    """
    pmns = 'vllm.bench.'
    for metric in ['request_rate', 'burstiness', 'max_concurrency',
                   'duration', 'completed', 'num_prompts',
                   'total_input_tokens', 'total_output_tokens',
                   'request_throughput', 'request_goodput',
                   'output_throughput', 'total_token_throughput',
                   'max_output_tokens_per_s', 'max_concurrent_requests',
                   'mean_ttft_ms', 'median_ttft_ms',
                   'std_ttft_ms', 'p99_ttft_ms',
                   'mean_tpot_ms', 'median_tpot_ms',
                   'std_tpot_ms', 'p99_tpot_ms',
                   'mean_itl_ms', 'median_itl_ms',
                   'std_itl_ms', 'p99_itl_ms',
                   'mean_e2el_ms', 'median_e2el_ms',
                   'std_e2el_ms', 'p99_e2el_ms']:
        if metric in values and values[metric] is not None:
            if args.verbose:
                print('adding', metric, ':', str(values[metric]))
            pcp.pmiPutValue(pmns + metric, instance, str(values[metric]))


def timestamp(timestring) -> datetime:
    # extract timestamp from string like "20250929-080417"
    (dates, times) = timestring.split('-')
    # (year, month, day, hour, minute, second, weekday, day_of_year, is_dst)
    (yy, mm, dd) = (int(dates[0:4]), int(dates[4:6]), int(dates[6:8]))
    (HH, MM, SS) = (int(times[0:2]), int(times[2:4]), int(times[4:6]))
    local_time_tuple = (yy, mm, dd, HH, MM, SS, 0, 0, 0)
    seconds_since_epoch = calendar.timegm(local_time_tuple)
    return datetime.fromtimestamp(seconds_since_epoch, timezone.utc)


parser = argparse.ArgumentParser()
parser.add_argument("-v", "--verbose", help="Enable verbose progress diagnostics",
                    action="store_true")
parser.add_argument("-a", "--archive", help="Performance Co-Pilot archive output (timestamp if not specified)")
parser.add_argument("-H", "--hostname", help="Performance Co-Pilot archive hostname (backend if not specified)")
parser.add_argument("results", help="JSON results from vllm or Model Furnace")
args = parser.parse_args()

with open(args.results) as json_data:
    try:
        payload = json.load(json_data)
    except: # pylint: disable=bare-except
        print('Failed to load JSON document from:', args.results)
        sys.exit(1)
    instid = 0
    backend = None

    # check if the results are wrapped with model furnace metadata
    try:
        result = payload['report'] # may generate a KeyError
    except KeyError:
        result = payload # default vllm bench format

    benchmark = result
    backend = benchmark['backend']
    if args.verbose:
        print('Backend', backend)
    # Create a new PCP archive
    if not args.hostname:
        args.hostname = backend
    if not args.archive:
        args.archive = timestamp(benchmark['date'])
        args.archive = args.archive.strftime('%Y%m%d.%H.%M')
    log = pmi.pmiLogImport(args.archive)
    log.pmiSetHostname(args.hostname)
    log.pmiSetTimezone('UTC')
    # extract some Model Furnace metadata as labels, when present
    for label in ['model_id', 'tokenizer_id', 'endpoint_type', 'backend']:
        value = payload.get(label)
        if value:
            log.pmiPutLabel(PM_LABEL_CONTEXT, 0, 0, label, value)
    ids = register_metrics(log)

    instid = instid + 1
    instname = benchmark['backend'] + '-' + benchmark['model_id']
    if args.verbose:
        print('[%d] %s' % (instid, instname))

    log.pmiAddInstance(ids, instname, instid)

    start_time = timestamp(benchmark['date'])
    put_metric_values(log, benchmark, instname)
    log.pmiWrite(start_time)

    finish_time = start_time + timedelta(seconds=benchmark['duration'])
    put_metric_values(log, benchmark, instname)
    log.pmiWrite(finish_time)

    log.pmiPutMark() # end of benchmark iteration

    if args.verbose:
        print("Writing archive:", args.archive)
    del log # flush the archive to persistent storage



#!/usr/bin/pmpython
#
# Copyright (C) 2022-2023,2025 Red Hat.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# pylint: disable=broad-except, bare-except, protected-access
""" Display geographical location from local cache or IP lookup """

from pcp.pmapi import pmContext as PCP
import urllib.request as httprequest
import urllib.error as httperror
import threading
import time
import json
import sys
import os

start = time.time()
output = '{"latitude":%s,"longitude":%s}'
writing = False
stdoutfd = sys.stdout

# first extract from cached (labels) file
try:
    if len(sys.argv) == 2:  # use single argument if presented
        path = sys.argv[1]
    else:
        path = PCP.pmGetConfig('PCP_SYSCONF_DIR')
        path += '/labels/optional/geolocate'
    cached = json.load(path)
    print(output % (cached['latitude'], cached['longitude']))
except:
    try:
        sys.stdout = open(path, 'w')    # create (or overwrite)
        writing = True
    except PermissionError:
        sys.stderr.write('No permission to write to %s\n' % path)
        sys.stdout = stdoutfd


# setup threads for handling parallel REST API requests
threads = []

def finish(url):
    if writing:
        sys.stdout.close()
        sys.stdout = stdoutfd
    #print("'%s\' fetched in %ss" % (url, time.time() - start))
    os._exit(0) # not sys.exit, which awaits all threads


def ipinfo(url):
    try:
        with httprequest.urlopen(url) as http:
            js = http.read().decode('utf-8')
            data = json.loads(js)
            coords = data['loc'].split(',')
            print(output % (coords[0], coords[1]))
        finish(url)
    except httperror.HTTPError:
        pass

IPINFO = "https://ipinfo.io/json"
threads.append(threading.Thread(target = ipinfo, args = (IPINFO, )))


try:
    for thread in threads:
        thread.start()
        thread.join()
except KeyboardInterrupt:
    finish('(none)')



#!/usr/bin/env python3
import os
import sys
import argparse
import json
from typing import IO


def check_trend(history_trend_path: IO):
    history = json.load(history_trend_path)
    if len(history) < 2:
        print("At least two test runs are required to see any trends.")
        return

    current = history[0]["data"]
    previous = history[1]["data"]

    print("Current test results, compared to the previous test run:\n")

    fmt = lambda x: f"{x:+5d}".replace("+0", " 0")
    print(f"Failed tests:  {current['failed']:5d} ({fmt(current['failed']-previous['failed'])})")
    print(f"Broken tests:  {current['broken']:5d} ({fmt(current['broken']-previous['broken'])})")
    print(f"Skipped tests: {current['skipped']:5d} ({fmt(current['skipped']-previous['skipped'])})")
    print(f"Passed tests:  {current['passed']:5d} ({fmt(current['passed']-previous['passed'])})")
    print()

    failures_diff = (current["failed"] + current["broken"]) - (previous["failed"] + previous["broken"])
    if failures_diff > 0:
        if "GITHUB_ACTION" in os.environ:
            print("::error::", end="")
        print(f"This commit contains {failures_diff} regression(s).")
        sys.exit(1)
    elif failures_diff == 0:
        if "GITHUB_ACTION" in os.environ:
            print("::notice::", end="")
        print("This commit contains no regressions.")
    else:
        if "GITHUB_ACTION" in os.environ:
            print("::notice::", end="")
        print(f"This commit resolves {-failures_diff} test failure(s).")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--history-trend", type=argparse.FileType("r"), required=True)
    args = parser.parse_args()

    check_trend(args.history_trend)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import os
import shutil
import argparse
import re
import json
import csv
from enum import Enum
from collections import defaultdict
from typing import IO, List, Dict, Optional, Tuple
import requests


class Test:
    class Status(str, Enum):
        Passed = "passed"
        Skipped = "skipped"
        Failed = "failed"
        Broken = "broken"
        Triaged = "triaged"

    def __init__(self, name: str):
        self.name = name
        self.groups = []
        self.platform = ""
        self.path = ""
        self.description = ""
        self.start = 0
        self.stop = 0
        self.status = ""
        self.message = ""


FAILED_TEST_STATES = [Test.Status.Failed, Test.Status.Broken, Test.Status.Triaged]


def read_test_durations(timings_path: str) -> Dict[str, List[int]]:
    test_durations = {}
    with open(timings_path, encoding="utf-8") as f:
        for line in f:
            no, start, stop = line.strip().split()
            test_durations[no] = [int(start), int(stop)]
    return test_durations


def read_test_groups(group_path: str) -> Dict[str, List[str]]:
    tests = {}
    with open(group_path, encoding="utf-8") as f:
        for line in f:
            spl = line.rstrip().split()
            if len(spl) >= 2 and spl[0].isdigit():
                tests[spl[0]] = spl[1:]
    return tests


def read_testlog(
    qa_dir: str,
    testartifacts_dir: str,
    groups: Dict[str, List[str]],
    test_durations: Dict[str, List[int]],
    platform: str,
) -> List[Test]:
    tests: List[Test] = []
    testlog_path = os.path.join(testartifacts_dir, "test.log")
    with open(testlog_path, encoding="utf-8", errors="backslashreplace") as testlog_file:
        for line in testlog_file:
            # [xx%] will be displayed if there are more than 9 tests
            #  Xs ... will be displayed if test was already run
            success_m = re.match(r"^(?:\[\d+%\] )?(\d+)(?: \d+s \.\.\.)?\n$", line)
            notrun_m = re.match(r"^(?:\[\d+%\] )?(\d+)(?: \d+s \.\.\.)? \[not run\] (.+)\n$", line)
            failed_m = re.match(r"^(?:\[\d+%\] )?(\d+)(?: \d+s \.\.\.)? ((\- |\[).+)\n$", line)
            cancelled_m = re.match(r"^(?:\[\d+%\] )?(\d+)(?: \d+s \.\.\.)?\Z", line)

            if success_m or notrun_m or failed_m or cancelled_m:
                test_no = (success_m or notrun_m or failed_m or cancelled_m).group(1)

                test = Test(test_no)
                test.groups = groups.get(test_no, [])
                test.platform = platform
                test.path = testartifacts_dir
                test.start, test.stop = test_durations.get(test_no, (0, 0))
                test.description = ""
                try:
                    with open(os.path.join(qa_dir, test_no), encoding="utf-8") as f:
                        next(f)  # skip shebang
                        for test_line in f:
                            if test_line == "\n":
                                break
                            test.description += test_line.strip(" #") + "\n"
                    tests.append(test)
                except IOError:  # non-existant file
                    # false match - it's probably output from the previous test
                    if tests and tests[-1].status in FAILED_TEST_STATES:
                        if len(tests[-1].message) < 10000:  # ignore output for huge diffs
                            tests[-1].message += line
                    continue

            if success_m:
                test.status = Test.Status.Passed
            elif notrun_m:
                test.status = Test.Status.Skipped
                test.message = notrun_m.group(2)
            elif failed_m:
                failed_msg = failed_m.group(2)
                if failed_msg.startswith("- "):
                    failed_msg = failed_msg[2:]

                if "failed" in failed_msg:
                    test.status = Test.Status.Broken
                elif "triaged" in failed_msg:
                    test.status = Test.Status.Triaged
                else:
                    test.status = Test.Status.Failed
                test.message = failed_msg + "\n"
            elif cancelled_m:
                test.status = Test.Status.Broken
                test.message = "test cancelled"
            elif tests and tests[-1].status in FAILED_TEST_STATES:
                # if this line doesn't match any regex, it's probably output from the previous test
                if len(tests[-1].message) < 10000:  # ignore output for huge diffs
                    tests[-1].message += line

    return tests


def read_platforms(qa_dir: str, artifacts_path: str) -> Tuple[List[str], List[Test]]:
    groups = read_test_groups(os.path.join(qa_dir, "group"))
    platforms = set()
    tests = []
    for artifact_dir in os.listdir(artifacts_path):
        if artifact_dir.startswith("build-"):
            platform = artifact_dir[len("build-") :]
            platforms.add(platform)
        elif artifact_dir.startswith("test-"):
            platform = artifact_dir[len("test-") :]
            platforms.add(platform)

            testartifacts_dir = os.path.join(artifacts_path, artifact_dir)
            test_timings_file = os.path.join(testartifacts_dir, "check.timings")

            test_durations = read_test_durations(test_timings_file)
            tests.extend(read_testlog(qa_dir, testartifacts_dir, groups, test_durations, platform))
    return list(platforms), tests


def write_test_report_csv(tests: List[Test], out: IO):
    writer = csv.writer(out)
    writer.writerow(["Name", "Groups", "Platform", "Start", "Stop", "Duration", "Status"])
    for test in tests:
        writer.writerow(
            [
                test.name,
                ",".join(test.groups),
                test.platform,
                test.start,
                test.stop,
                test.stop - test.start,
                test.status,
            ]
        )


def test_summary_short_platform(platform: str) -> str:
    return (
        platform.replace("-container", "")
        .replace("fedora", "f")
        .replace("centos-stream", "el")
        .replace("centos", "el")
        .replace("debian", "deb")
        .replace("ubuntu", "ubu")
    )


def test_summary_tests(tests: List[Test]):
    tests_grouped = defaultdict(dict)
    platform_set = set()
    has_failed_tests = False
    for test in tests:
        tests_grouped[test.name][test.platform] = test
        platform_set.add(test.platform)
        if not has_failed_tests and test.status in FAILED_TEST_STATES:
            has_failed_tests = True

    if not has_failed_tests:
        return ""

    platforms = sorted(platform_set)
    platform_short = {platform: test_summary_short_platform(platform) for platform in platforms}

    summary = "=== Failed Tests ===\n\n"
    summary += " Test "
    for platform in platforms:
        summary += f" {platform_short[platform]}"
    summary += "  Groups\n"

    for test_name in sorted(tests_grouped, key=int):
        test_line = f" {int(test_name):4d} "
        for platform in platforms:
            col_width = len(platform_short[platform])
            test_line += " "

            test = tests_grouped[test_name].get(platform)
            if test and test.status == Test.Status.Passed:
                test_line += " ".center(col_width)
            elif test and test.status in [Test.Status.Failed, Test.Status.Broken]:
                test_line += "X".center(col_width)
            elif test and test.status == Test.Status.Triaged:
                test_line += "T".center(col_width)
            else:
                test_line += "-".center(col_width)

        groups = sorted(list(tests_grouped[test_name].values())[0].groups)
        test_line += f"  {' '.join(groups)}"

        # skip tests without any failures
        if "X" in test_line or "T" in test_line:
            summary += test_line + "\n"

    summary += "\nLegend: ( ) Passed, (X) Failure, (T) Triaged, (-) Skipped\n\n\n"
    return summary


def test_summary_platforms(platforms: List[str], tests: List[Test]):
    platform_stats = {platform: {"passed": 0, "failed": 0, "skipped": 0, "cancelled": 0} for platform in platforms}
    for test in tests:
        if test.status == Test.Status.Passed:
            platform_stats[test.platform]["passed"] += 1
        elif test.status == Test.Status.Skipped:
            platform_stats[test.platform]["skipped"] += 1
        elif test.status == Test.Status.Broken and test.message == "test cancelled":
            platform_stats[test.platform]["cancelled"] += 1
        elif test.status in FAILED_TEST_STATES:
            platform_stats[test.platform]["failed"] += 1

    summary = "=== Platform Summary ===\n\n"
    summary += f"{'Platform':25}  Pass Fail Skip\n"
    for platform, stats in sorted(platform_stats.items()):
        summary += f"{platform:25}  {stats['passed']:4d} {stats['failed']:4d} {stats['skipped']:4d}"
        if stats["passed"] == 0:
            summary += "  build broken\n"
        elif stats["cancelled"] > 0:
            summary += "  QA ran into a timeout\n"
        elif stats["failed"] == 0:
            summary += "  ✓\n"
        else:
            summary += "\n"
    summary += "\n\n"
    return summary


def write_test_summary(platforms: List[str], tests: List[Test], out: IO, build_url: str, report_url: str):
    summary = test_summary_tests(tests)
    summary += test_summary_platforms(platforms, tests)
    summary += f"Build:  {build_url}\n"
    summary += f"Report: {report_url}\n"
    out.write(summary)


def write_allure_test_result(test: Test, allure_results_path: str, source_url: Optional[str]):
    allure_result = {
        "name": f"QA {test.name}",
        "fullName": f"QA #{test.name} on {test.platform}",
        "historyId": f"{test.name}@{test.platform}",
        "description": test.description,
        # Allure doesn't have a "triaged" test status, therefore we use the "unknown" status
        "status": "unknown" if test.status == Test.Status.Triaged else test.status,
        "statusDetails": {"message": test.message, "flaky": "flakey" in test.groups},
        "attachments": [],
        "start": test.start * 1000,
        "stop": test.stop * 1000,
        "labels": [
            {"name": "suite", "value": test.platform},
            {"name": "host", "value": test.platform},
        ],
        "links": [
            {
                "type": "test_case",
                "url": f"{source_url}/qa/{test.name}" if source_url else "",
                "name": "Source",
            },
            {
                "type": "test_case",
                "url": f"{source_url}/qa/{test.name}.out" if source_url else "",
                "name": "Expected output",
            },
        ],
    }

    for group in test.groups:
        allure_result["labels"].append({"name": "epic", "value": group})

    if test.status in FAILED_TEST_STATES:
        allure_result["description"] += (
            f"**To reproduce this test, please run:**\n\n"
            f"    build/ci/ci-run.py {test.platform} reproduce"
        )

        out_bad = os.path.join(test.path, f"{test.name}.out.bad")
        if os.path.exists(out_bad):
            out_bad_attachment_fn = f"{test.platform}-{test.name}.out.bad-attachment.txt"
            shutil.copyfile(out_bad, os.path.join(allure_results_path, out_bad_attachment_fn))
            allure_result["attachments"].append(
                {
                    "type": "text/plain",
                    "name": f"{test.name}.out.bad",
                    "source": out_bad_attachment_fn,
                }
            )

        out_full = os.path.join(test.path, f"{test.name}.full")
        if os.path.exists(out_full):
            out_full_attachment_fn = f"{test.platform}-{test.name}.full-attachment.txt"
            shutil.copyfile(out_full, os.path.join(allure_results_path, out_full_attachment_fn))
            allure_result["attachments"].append(
                {
                    "type": "text/plain",
                    "name": f"{test.name}.full",
                    "source": out_full_attachment_fn,
                }
            )

    with open(os.path.join(allure_results_path, f"{test.platform}-{test.name}-result.json"), "w") as f:
        json.dump(allure_result, f)


# pylint: disable=too-many-arguments
def write_allure_results(
    allure_results_path: str,
    tests: List[Test],
    source_url: Optional[str],
    build_name: Optional[str],
    build_url: Optional[str],
    report_url: Optional[str],
):
    os.makedirs(allure_results_path, exist_ok=True)
    for test in tests:
        write_allure_test_result(test, allure_results_path, source_url)

    executor = {
        "name": "GitHub Actions",
        "type": "github",
        "buildName": build_name,
        "buildUrl": build_url,
        "reportUrl": report_url,
    }
    with open(os.path.join(allure_results_path, "executor.json"), "w") as f:
        json.dump(executor, f)


def send_slack_notification(
    platforms: List[str],
    tests: List[Test],
    slack_channel: str,
    build_url: str,
    report_url: str,
    summary_url: str,
):
    platform_stats = defaultdict(lambda: {"passed": 0, "failed": 0, "skipped": 0, "cancelled": 0})
    for test in tests:
        if test.status == Test.Status.Passed:
            platform_stats[test.platform]["passed"] += 1
        elif test.status == Test.Status.Skipped:
            platform_stats[test.platform]["skipped"] += 1
        elif test.status == Test.Status.Broken and test.message == "test cancelled":
            platform_stats[test.platform]["cancelled"] += 1
        elif test.status in FAILED_TEST_STATES:
            platform_stats[test.platform]["failed"] += 1

    platform_texts = []
    for platform in sorted(platforms):
        if platform in platform_stats:
            stats = platform_stats[platform]
            platform_stats_text = f"Passed: {stats['passed']}, Failed: {stats['failed']}, Skipped: {stats['skipped']}"
            if stats["cancelled"] > 0:
                platform_stats_text = f"CANCELLED ({platform_stats_text})"

            if stats["cancelled"] > 0 or stats["passed"] < 1000:
                symbol = ":x:"
            elif stats["failed"] > 0:
                symbol = ":warning:"
            else:
                symbol = ":heavy_check_mark:"
        else:
            platform_stats_text = "Build BROKEN"
            symbol = ":x:"
        platform_texts.append(f"*{platform}*:\n{symbol} {platform_stats_text}")

    platform_blocks = []
    # slack only allows max 10 fields inside one block
    for i in range(0, len(platform_texts), 10):
        platform_blocks.append(
            {
                "type": "section",
                "fields": [{"type": "mrkdwn", "text": text} for text in platform_texts[i : i + 10]],
            }
        )

    payload = {
        "channel": slack_channel,
        "attachments": [
            {
                "blocks": [
                    {
                        "type": "section",
                        "text": {"type": "mrkdwn", "text": "*Daily QA:*"},
                    }
                ]
                + platform_blocks
                + [
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f"*<{build_url}|View build logs>*\n"
                            f"*<{report_url}|View QA report>*\n"
                            f"*<{summary_url}|View test summary>*",
                        },
                    }
                ]
            }
        ],
    }

    print("Sending slack notification...")
    r = requests.post(
        "https://slack.com/api/chat.postMessage",
        headers={"Authorization": f"Bearer {os.environ['SLACK_BOT_TOKEN']}"},
        json=payload,
    )
    print(r.text)
    r.raise_for_status()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--qa", default="./qa")
    parser.add_argument("--artifacts")
    parser.add_argument("--summary", type=argparse.FileType("w"))
    parser.add_argument("--csv", type=argparse.FileType("w"))
    parser.add_argument("--allure-results", dest="allure_results")
    parser.add_argument("--source-url", dest="source_url")  # require for allure report
    parser.add_argument("--build-name", dest="build_name")  # require for allure report
    parser.add_argument("--build-url", dest="build_url")  # required for allure report and slack message
    parser.add_argument("--report-url", dest="report_url")  # required for allure report and slack message
    parser.add_argument("--slack-channel", dest="slack_channel")  # required for slack message
    parser.add_argument("--summary-url", dest="summary_url")  # required for slack message
    args = parser.parse_args()

    platforms, tests = read_platforms(args.qa, args.artifacts)

    if args.summary:
        write_test_summary(platforms, tests, args.summary, args.build_url, args.report_url)

    if args.csv:
        write_test_report_csv(tests, args.csv)

    if args.allure_results:
        write_allure_results(
            args.allure_results, tests, args.source_url, args.build_name, args.build_url, args.report_url
        )

    if args.slack_channel:
        print()
        send_slack_notification(platforms, tests, args.slack_channel, args.build_url, args.report_url, args.summary_url)


if __name__ == "__main__":
    main()



#!/usr/bin/env pmpython
#
# Copyright (c) 2023 Oracle and/or its affiliates.
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# pylint: disable=bad-whitespace,too-many-arguments,too-many-lines, bad-continuation
# pylint: disable=redefined-outer-name,unnecessary-lambda
#
import signal
import sys
import time
from pcp import pmapi, pmcc
from cpmapi import PM_CONTEXT_ARCHIVE, PM_MODE_FORW

SYS_METRICS= ["kernel.uname.sysname","kernel.uname.release",
               "kernel.uname.nodename","kernel.uname.machine","hinv.ncpu"]
BUDDYSTAT_METRICS = ["mem.buddyinfo.pages","mem.buddyinfo.total"]

ALL_METRICS = BUDDYSTAT_METRICS + SYS_METRICS

def adjust_length(name,size):
    return name.ljust(size)

class ReportingMetricRepository:

    def __init__(self,group):
        self.group=group
        self.current_cached_values = {}

    def __fetch_current_value(self,metric):
        val=dict(map(lambda x: (x[1], x[2]), self.group[metric].netValues))
        return dict(val)

    def current_value(self,metric):
        if metric not in self.group:
            return None
        if self.current_cached_values.get(metric) is None:
            first_value=self.__fetch_current_value(metric)
            self.current_cached_values[metric]=first_value
        return self.current_cached_values[metric]

class BuddyStatUtil:
    def __init__(self,metrics_repository):
        self.__metric_repository=metrics_repository
        self.report=ReportingMetricRepository(self.__metric_repository)

    def buddy_pages(self):
        return self.report.current_value('mem.buddyinfo.pages')

    def buddy_total_pages(self):
        return self.report.current_value('mem.buddyinfo.total')

    def names(self):
        data = self.report.current_value('mem.buddyinfo.pages')
        return data.keys()

class BuddyinfoReport(pmcc.MetricGroupPrinter):
    def __init__(self,opts,group):
        self.opts=opts
        self.group=group
        self.context=opts.context
        self.samples=opts.samples
        self.header = "unknown"

    def __get_ncpu(self, group):
        return group['hinv.ncpu'].netValues[0][2]

    def __print_machine_info(self, context):
        if self.header == "unknown":
            try:
                timestamp = self.group.pmLocaltime(context.timestamp.tv_sec)
                # Please check strftime(3) for different formatting options.
                # Also check TZ and LC_TIME environment variables for more
                    # information on how to override the default formatting of
                    # the date display in the header
                time_string = time.strftime("%x", timestamp.struct_time())
                header_string = ''
                header_string += context['kernel.uname.sysname'].netValues[0][2] + '  '
                header_string += context['kernel.uname.release'].netValues[0][2] + '  '
                header_string += '(' + context['kernel.uname.nodename'].netValues[0][2] + ')  '
                header_string += time_string + '  '
                header_string += context['kernel.uname.machine'].netValues[0][2] + '  '
                self.header = header_string
            except  IndexError:
                pass
        try:
            print("%s  (%s CPU)" % (self.header, self.__get_ncpu(context)))
        except  IndexError:
            pass

    def __print_header(self,header_indentation,value_indentation):
        value_indentation+=" "*2
        print("TimeStamp"+ " "+header_indentation + "Normal" + header_indentation+value_indentation+" " +
              "Nodes" + header_indentation + "Order0" + value_indentation +
              "Order1" + value_indentation +"Order2" + value_indentation + "Order3" + value_indentation +
              "Order4" + value_indentation +"Order5" + value_indentation + "Order6" + value_indentation +
              "Order7" + value_indentation +"Order8" + value_indentation + "Order9" +
              value_indentation + "Order10")

    def __print_values(self,timestamp,header_indentation,\
                     value_indentation,buddystatus):
        names=buddystatus.names()
        pages=buddystatus.buddy_pages()
        order_set = set()
        nodes_set = set()
        no_of_nodes_set = set()
        for name in names:
            part=name.split('::')
            if len(part)==3:
                nodes_set.add(part[0])
                order_set.add(part[1])
                no_of_nodes_set.add(part[2])
        def __extract_numeric_part(element):
            return int(element[5:])
        for Normal in sorted(nodes_set):
            for node in no_of_nodes_set:
                value=""
                for order in sorted(order_set,key=__extract_numeric_part):
                    nodename = adjust_length(Normal,9) if len(Normal) < 9 else Normal
                    key = "%s::%s::%s" % (Normal, order, node)
                    data = str(pages.get(key,0))
                    value +=  adjust_length(data,8) if len(data) < 8 else data
                    value += value_indentation

                print("%s %s %s %s %s %s %s "%(timestamp,header_indentation,nodename,header_indentation,node,
                                                  header_indentation,value))

    def print_report(self,group,timestamp,header_indentation,value_indentation):

        def __print_buddy_status():
            buddystatus = BuddyStatUtil(group)
            if buddystatus.names():
                try:
                    self.__print_machine_info(group)
                    self.__print_header(header_indentation, value_indentation)
                    self.__print_values(timestamp, header_indentation, value_indentation, buddystatus)
                except IndexError:
                    print("Incorrect machine info due to some missing metrics")
                return
            else:
                return

        if self.context != PM_CONTEXT_ARCHIVE and self.samples is None:
            __print_buddy_status()
            sys.exit(0)
        elif self.context == PM_CONTEXT_ARCHIVE and self.samples is None:
            __print_buddy_status()
        elif self.samples >=1:
            __print_buddy_status()
            self.samples-=1

    def report(self, manager):
        group = manager["allinfo"]
        self.samples=self.opts.pmGetOptionSamples()
        t_s = group.contextCache.pmLocaltime(int(group.timestamp))
        timestamp = time.strftime(BuddyinfoOptions.timefmt, t_s.struct_time())
        header_indentation = "        " if len(timestamp) < 9 else (len(timestamp) - 7) * " "
        value_indentation = ((len(header_indentation) + 2) - len(timestamp)) * " "
        self.print_report(group,timestamp,header_indentation,value_indentation)

class BuddyinfoOptions(pmapi.pmOptions):
    timefmt = "%H:%M:%S"
    uflag = False

    def extraOptions(self,opt,optarg,index):
        if opt == 'u':
            BuddyinfoOptions.uflag = True

    def __init__(self):
        pmapi.pmOptions.__init__(self, "a:s:Z:uzV?")
        self.pmSetLongOptionHeader("General options")
        self.pmSetOptionCallback(self.extraOptions)
        self.pmSetLongOptionArchive()
        self.pmSetLongOption("no-interpol", 0, "u", "", "disable interpolation mode with archives")
        self.pmSetLongOptionHostZone()
        self.pmSetLongOptionTimeZone()
        self.pmSetLongOptionHelp()
        self.pmSetLongOptionSamples()
        self.pmSetLongOptionVersion()
        self.samples=None
        self.context=None
    def checkOptions(self, manager):
        if BuddyinfoOptions.uflag:
            if manager._options.pmGetOptionInterval():  # pylint: disable=protected-access
                print("Error: -t incompatible with -u")
                return False
            if manager.type != PM_CONTEXT_ARCHIVE:
                print("Error: -u can only be specified with -a archive")
                return False
        return True

if __name__ == '__main__':
    try:
        opts = BuddyinfoOptions()
        mngr = pmcc.MetricGroupManager.builder(opts,sys.argv)
        opts.context=mngr.type

        if not opts.checkOptions(mngr):
            raise pmapi.pmUsageErr
        if BuddyinfoOptions.uflag:
            # -u turns off interpolation
            mngr.pmSetMode(PM_MODE_FORW, mngr._options.pmGetOptionOrigin(), None) # pylint: disable=protected-access

        missing = mngr.checkMissingMetrics(ALL_METRICS)
        if missing is not None:
            sys.stderr.write('Error: not all required metrics are available\nMissing: %s\n' % (missing))
            sys.exit(1)
        mngr["buddyinfo"] = BUDDYSTAT_METRICS
        mngr["allinfo"]=ALL_METRICS
        mngr.printer = BuddyinfoReport(opts,mngr)
        sts = mngr.run()
        sys.exit(sts)
    except pmapi.pmErr as error:
        sys.stderr.write('%s\n' % (error.message()))
    except pmapi.pmUsageErr as usage:
        usage.message()
        sys.exit(1)
    except IOError:
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    except KeyboardInterrupt:
        pass



#!/usr/bin/env pmpython
#
# Copyright (c) 2022 Oracle and/or its affiliates.
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from mock import Mock
import unittest
from pcp_ps import ProcessStatusReporter


class TestProcessStateReporter(unittest.TestCase):
    def setUp(self):
        self.options = Mock(
            show_process_user=None)

        process_1 = Mock(pid=Mock(return_value=1),
                         process_name=Mock(return_value="process_1"),
                         user_name=Mock(return_value='pcp'),
                         user_id=Mock(return_value=1000),
                         stack_size=Mock(return_value=136),
                         tty_name=Mock(return_value="tty"),
                         total_time=Mock(return_value=100))

        self.processes = [process_1]

    def test_print_report_with_user_name(self):
        self.options.show_all_process = True
        self.options.print_count=1
        process_stack_util = Mock()
        process_filter = Mock()
        printer = Mock()
        process_filter.filter_processes = Mock(return_value=self.processes)
        reporter = ProcessStatusReporter(process_stack_util, process_filter, 1.34, printer, self.options)

        reporter.print_report(123, "  ", "    ")

        printer.assert_called_with("123    1\t\ttty\t100\tprocess_1")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env pmpython
#
# Copyright (c) 2022 Oracle and/or its affiliates.
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# pylint: disable=bad-whitespace,too-many-lines
# pylint: disable=too-many-arguments,too-many-positional-arguments
# pylint: disable=redefined-outer-name,unnecessary-lambda
#


import sys
import time
import signal
from pcp import pmcc
from pcp import pmapi
from cpmapi import PM_CONTEXT_ARCHIVE

process_state_info = {}

PSSTAT_METRICS = ['kernel.uname.nodename', 'kernel.uname.release', 'kernel.uname.sysname',
                  'kernel.uname.machine', 'hinv.ncpu', 'proc.psinfo.pid', 'proc.psinfo.guest_time',
                  'proc.psinfo.utime', 'proc.psinfo.ppid', 'proc.psinfo.rt_priority', 'proc.psinfo.rss',
                  'proc.id.uid_nm', 'proc.psinfo.stime', 'kernel.all.boottime', 'proc.psinfo.sname',
                  'proc.psinfo.start_time', 'proc.psinfo.vsize', 'proc.psinfo.priority',
                  'proc.psinfo.nice', 'proc.psinfo.wchan_s', 'proc.psinfo.psargs', 'proc.psinfo.cmd',
                  'proc.psinfo.ttyname', 'mem.physmem', 'proc.psinfo.policy']

SCHED_POLICY = ['NORMAL', 'FIFO', 'RR', 'BATCH', '', 'IDLE', 'DEADLINE']


class StdoutPrinter:
    def Print(self, args):
        print(args)


class NoneHandlingPrinterDecorator:
    def __init__(self, printer):
        self.printer = printer

    def Print(self, args):
        new_args = args.replace('None', '?')
        self.printer.Print(new_args)


# After fetching non singular metric values, create a mapping of instance id
# to instance value rather than instance name to instance value.
# The reason is, in PCP, instance names require a separate pmGetIndom() request
# and some of the names may not be available.
class ReportingMetricRepository:
    def __init__(self, group):
        self.group = group
        self._current_cache = {}
        self._previous_cache = {}

    def _fetch_values(self, metric, use_previous=False):
        """Fetch values - always returns a dictionary."""
        if metric not in self.group:
            return {}
        attr = "netPrevValues" if use_previous else "netValues"
        values = getattr(self.group[metric], attr, [])
        return {x[0].inst: x[2] for x in values} if values else {}

    def _get_values_dict(self, metric, use_previous=False):
        """Get cached dictionary of all values for a metric."""
        cache = self._previous_cache if use_previous else self._current_cache
        if metric not in cache:
            cache[metric] = self._fetch_values(metric, use_previous)
        return cache[metric]

    def current_value(self, metric, instance=None):
        """Get current value. Returns single value if instance given, else returns dict."""
        values_dict = self._get_values_dict(metric, use_previous=False)
        if instance is not None:
            return values_dict.get(instance)
        return values_dict

    def previous_value(self, metric, instance=None):
        """Get previous value. Returns single value if instance given, else returns dict."""
        values_dict = self._get_values_dict(metric, use_previous=True)
        if instance is not None:
            return values_dict.get(instance)
        return values_dict

    def current_values(self, metric_name):
        """Get all current values for a metric (returns dict)."""
        return self._get_values_dict(metric_name, use_previous=False)

    def previous_values(self, metric_name):
        """Get all previous values for a metric (returns dict)."""
        return self._get_values_dict(metric_name, use_previous=True)


class ProcessFilter:
    """
    Optimized filtering of processes based on user-provided options.
    Precomputes filter predicates at instantiation to minimize overhead per process,
    and leverages generator expressions with all() for efficiency.
    """
    def __init__(self, options):
        self.options = options
        self._predicates = []
        flag = getattr(options, "universal_flag", None)
        # Username filter
        if flag == "username" and getattr(options, "filtered_process_user", None) is not None:
            required_user = options.filtered_process_user.strip()
            self._predicates.append(lambda proc: proc.user_name().strip() == required_user)
        # PID filter
        if flag == "pid" and getattr(options, "pid_list", None) is not None:
            try:
                pids = set(int(pid) for pid in options.pid_list)
                self._predicates.append(lambda proc: int(proc.pid()) in pids)
            except Exception:
                self._predicates.append(lambda proc: False)
        # PPID filter
        if flag == "ppid" and getattr(options, "ppid_list", None) is not None:
            try:
                ppids = set(int(ppid) for ppid in options.ppid_list)
                self._predicates.append(lambda proc: int(proc.ppid()) in ppids)
            except Exception:
                self._predicates.append(lambda proc: False)
        # Command name filter
        if flag == "command" and getattr(options, "command_list", None) is not None:
            commands = set(cmd.strip() for cmd in options.command_list if cmd is not None)
            self._predicates.append(lambda proc: (proc.process_name() or "").strip() in commands)
        # If no filter flag: no filter
        if not getattr(options, "filter_flag", False) or not self._predicates:
            self._predicates.append(lambda proc: True)

    def filter_processes(self, processes):
        """Yield processes matching all active predicates."""
        return (proc for proc in processes if all(pred(proc) for pred in self._predicates))

class ProcessStatusUtil:
    def __init__(self, instance, manager, delta_time, metrics_repository):
        self.instance = instance
        self.manager = manager
        self.__delta_time = delta_time
        self.__metric_repository = metrics_repository

    def __get_value(self, metric, instance=None):
        return self.__metric_repository.current_value(metric, instance)
    def __get_previous_value(self, metric, instance=None):
        return self.__metric_repository.previous_value(metric, instance)

    def pid(self):
        data = str(self.__get_value('proc.psinfo.pid', self.instance))
        return data.ljust(8) if len(data) < 8 else data


    def ppid(self):
        data = str(self.__get_value('proc.psinfo.ppid', self.instance))
        return data.ljust(8) if len(data) < 8 else data

    def user_name(self):
        data = self.__get_value('proc.id.uid_nm', self.instance)
        if data is None:
            return data
        return data.ljust(10) if len(data) < 10 else data[:10]

    def process_name(self):
        data = self.__get_value('proc.psinfo.cmd', self.instance)
        if data is None:
            return '-'
        return data.ljust(20) if len(data) < 20 else data[:20]

    def process_name_with_args(self, flag=False):
        data = self.__get_value('proc.psinfo.psargs', self.instance)
        if data is None:
            return data
        max_length = 30
        if not flag:
            return data[:max_length].ljust(max_length)
        else:
            return data

    def process_name_with_args_last(self):
        return self.process_name_with_args(True)
    def vsize(self):
        return self.__get_value('proc.psinfo.vsize', self.instance)

    def rss(self):
        return self.__get_value('proc.psinfo.rss', self.instance)
    def mem(self):
        total_mem = self.__get_value('mem.physmem', None)
        if isinstance(total_mem, dict):
            total_mem = next(iter(total_mem.values()), None)
        rss = self.__get_value('proc.psinfo.rss', self.instance)
        if total_mem is not None and rss is not None:
            return float("%.2f" % (100 * float(rss) / total_mem))
        else:
            return None

    def s_name(self):
        return self.__get_value('proc.psinfo.sname', self.instance)

    def cpu_number(self):
        return self.__get_value('proc.psinfo.processor', self.instance)

    def system_percent(self):
        c_systemtime = self.__get_value('proc.psinfo.stime', self.instance)
        p_systemtime = self.__get_previous_value('proc.psinfo.stime', self.instance)
        if c_systemtime is not None and p_systemtime is not None:
            system_time_diff = float(c_systemtime - p_systemtime)
            delta_time_in_ms = float(1000 * self.__delta_time)
            percent_of_time = 100 * system_time_diff / delta_time_in_ms
            return float("%.2f" % percent_of_time)
        else:
            return '-'

    def wchan_s(self):
        process = self.__get_value('proc.psinfo.wchan_s', self.instance)
        if process is None or process == "0" or process == '':
            return '-' + ' ' * 29
        return process.ljust(30) if len(process) < 30 else process[:30]

    def priority(self):
        return self.__get_value('proc.psinfo.priority', self.instance)

    def user_percent(self):
        c_usertime = self.__get_value('proc.psinfo.utime', self.instance)
        p_usertime = self.__get_previous_value('proc.psinfo.utime', self.instance)
        if c_usertime is not None and p_usertime is not None:
            percent_of_time = 100 * float(c_usertime - p_usertime) / float(1000 * self.__delta_time)
            return float("%.2f" % percent_of_time)
        else:
            return None

    def guest_percent(self):
        c_guesttime = self.__get_value('proc.psinfo.guest_time', self.instance)
        p_guesttime = self.__get_previous_value('proc.psinfo.guest_time', self.instance)
        if c_guesttime is not None and p_guesttime is not None:
            percent_of_time = 100 * float(c_guesttime - p_guesttime) / float(1000 * self.__delta_time)
            return float("%.2f" % percent_of_time)
        else:
            return None

    def total_percent(self):
        if self.user_percent() is not None and self.guest_percent() is not None and self.system_percent() is not None:
            return float("%.2f" % (self.user_percent() + self.guest_percent() + self.system_percent()))
        else:
            return 0.0

    def stime(self):
        c_systime = self.__get_value('proc.psinfo.stime', self.instance)
        p_systime = self.__get_previous_value('proc.psinfo.stime', self.instance)
        # sometimes the previous_value seems to be Nonetype, not sure why
        if p_systime is None:  # print a '?' here
            return '?'
        else:
            return c_systime - p_systime

    def start(self):
        s_time = self.__get_value('proc.psinfo.start_time', self.instance)
        group = self.manager['psstat']
        try:
            kernel_boottime = group['kernel.all.boottime'].netValues[0][2]
        except (KeyError, IndexError, TypeError, AttributeError):
            return '?'
        if s_time is None or kernel_boottime is None:
            return '?'
        ts_val = kernel_boottime + (s_time / 1000.0)
        try:
            ts = group.contextCache.pmLocaltime(int(ts_val))
        except Exception:
            return '?'
        if group.timestamp.tv_sec - ts_val >= 24*60*60:
            return time.strftime("%b%d %H:%M", ts.struct_time())
        else:
            return time.strftime("%H:%M:%S", ts.struct_time())


    def total_time(self):
        c_usertime = self.__get_value('proc.psinfo.stime', self.instance)
        p_guesttime = self.__get_previous_value('proc.psinfo.utime', self.instance)
        timefmt = "%H:%M:%S"
        if c_usertime and p_guesttime is not None:
            total_time = (c_usertime / 1000) + (p_guesttime / 1000)
        else:
            total_time = 0
        return time.strftime(timefmt, time.gmtime(total_time))

    def tty_name(self):
        return self.__get_value('proc.psinfo.ttyname', self.instance)

    def user_id(self):
        return self.__get_value('proc.id.uid', self.instance)

    def start_time(self):
        return self.__get_value('proc.psinfo.start_time', self.instance)

    def func_state(self):
        s_name = self.__get_value('proc.psinfo.sname', self.instance)
        if s_name == 'R':
            return '-'
        elif s_name is None:
            return '-'
        else:
            return self.wchan_s()

    def policy(self):
        policy_int = self.__get_value('proc.psinfo.policy', self.instance)
        if isinstance(policy_int, int) and 0 <= policy_int < len(SCHED_POLICY):
            return SCHED_POLICY[policy_int]
        return '?'


PIDINFO_PAIR = {"%cpu": ('%CPU', ProcessStatusUtil.system_percent),
                "%mem": ('%MEM', ProcessStatusUtil.mem),
                "start": ("START\t", ProcessStatusUtil.start),
                "time": ("TIME\t", ProcessStatusUtil.total_time),
                "cls": ("CLS", ProcessStatusUtil.policy),
                "cmd": ("Command\t\t", ProcessStatusUtil.process_name),
                "args": ("Command\t\t\t", ProcessStatusUtil.process_name_with_args),
                "args_last": ("Command", ProcessStatusUtil.process_name_with_args_last),
                "pid": ("PID\t", ProcessStatusUtil.pid),
                "ppid": ("PPID\t", ProcessStatusUtil.ppid),
                "pri": ("PRI", ProcessStatusUtil.priority),
                "state": ("S", ProcessStatusUtil.s_name),
                "rss": ("RSS", ProcessStatusUtil.rss),
                "rtprio": ("RTPRIO", ProcessStatusUtil.priority),
                "tty": ("TTY", ProcessStatusUtil.tty_name),
                "pname": ("Pname\t\t", ProcessStatusUtil.process_name),
                "vsize": ("VSZ", ProcessStatusUtil.vsize),
                "uname": ("USER\t", ProcessStatusUtil.user_name),
                "uid": ("USER_ID", ProcessStatusUtil.user_id),
                "wchan": ("WCHAN\t\t\t", ProcessStatusUtil.wchan_s)}


class ProcessStatus:
    def __init__(self, manager, metric_repository):
        self.__manager = manager
        self.__metric_repository = metric_repository

    def get_processes(self, delta_time):
        # Use generator expression for lazy evaluation and reduced memory usage
        return (ProcessStatusUtil(pid, self.__manager, delta_time, self.__metric_repository) for pid in self.__pids())

    def __pids(self):
        pid_dict = self.__metric_repository.current_values('proc.psinfo.pid')
        return sorted(pid_dict.values())


class DynamicProcessReporter:
    def __init__(self, process_report, process_filter, delta_time, printer, processStatOptions):
        self.delta_time = delta_time
        self.process_report = process_report
        self.process_filter = process_filter
        self.printer = printer
        self.processStatOptions = processStatOptions

    def __sort_by_idx(self, output_list, sorting_idx, reverse=True):
        # Rows are tab-delimited; splitting on whitespace breaks when
        # command/args contain spaces and shifts sortable column indexes.
        return sorted(
            output_list,
            key=lambda row: (
                float(row.split('\t')[sorting_idx].strip())
                if sorting_idx < len(row.split('\t'))
                and row.split('\t')[sorting_idx].strip().replace('.', '', 1).replace('-', '', 1).isdigit()
                else float('-inf')
            ),
            reverse=reverse
        )

    def _is_last_and_args(self, key):
        return (key == "args") and \
        self.processStatOptions.column_list.index(key) == len(self.processStatOptions.column_list) - 1
    def print_report(self, timestamp, header_indentation, value_indentation):

        # Exit logic for non-archive context
        if self.processStatOptions.context is not PM_CONTEXT_ARCHIVE and self.processStatOptions.print_count == 0:
            sys.exit(0)

        # Sorting validations
        sorting_idx = None
        if self.processStatOptions.sorting_flag:
            # For dynamic output, sorting key must be present in selected columns.
            if self.processStatOptions.sorting_order not in self.processStatOptions.column_list:
                raise ValueError("Sorting order not found in output columns")

            # Find sorting column index and adjust for Timestamp at position 0.
            sorting_idx = next((idx for idx, key in enumerate(self.processStatOptions.column_list)
                                if key == self.processStatOptions.sorting_order), None)
            # to account for Timestamp colum
            sorting_idx += 1

        # Always compute process list ONCE
        processes = self.process_filter.filter_processes(
            self.process_report.get_processes(self.delta_time)
        )

        output_list = []
        header = None

        # -------- Dynamic column list path -------- #
        if self.processStatOptions.column_list is not None:

            header = "Timestamp\t"
            for key in self.processStatOptions.column_list:
                if key in PIDINFO_PAIR:
                    header += PIDINFO_PAIR[key][0] + "\t"

            for process in processes:
                row = [timestamp]
                for key in self.processStatOptions.column_list:
                    if self._is_last_and_args(key):
                        row.append(str(PIDINFO_PAIR["args_last"][1](process)))
                    elif key in PIDINFO_PAIR:
                        row.append(str(PIDINFO_PAIR[key][1](process)))
                # print(row)
                output_list.append("\t".join(str(x) if x is not None else '' for x in row))

        # -------- Invalid column list -------- #
        # This should never happen, but just in case
        else:
            raise ValueError("No valid filterstate or column list provided")

        # Sorting logic
        if self.processStatOptions.sorting_flag and sorting_idx is not None:
            output_list = self.__sort_by_idx(output_list, sorting_idx, reverse=True)

        # --------- Print output  --------- #
        self.printer(header)
        self.printer('\n'.join(output_list))



class ProcessStatusReporter:
    def __init__(self, process_report, process_filter, delta_time, printer, processStatOptions):
        self.delta_time = delta_time
        self.process_report = process_report
        self.process_filter = process_filter
        self.printer = printer
        self.processStatOptions = processStatOptions

    def print_report(self, timestamp, header_indentation, value_indentation):
        if self.processStatOptions.debug_mode:
            print("option selected: %s" % self.processStatOptions.universal_flag)
            print("filer option status: %s" % self.processStatOptions.filterstate)
        FORMAT_MAP = {
            "empty_arg" : "Timestamp" + header_indentation + "PID\t\tTIME\t\tCMD",
            "all" : "Timestamp" + header_indentation + "PID\t\t\tTTY\tTIME\t\tCMD",
            "user": "Timestamp" + header_indentation + "USERNAME\tPID\t\t%CPU\t%MEM\tVSZ\tRSS\t" +
                    "TTY\tSTAT\tTIME\t\tSTART\t\tCOMMAND",
            "pid": "Timestamp" + header_indentation + "PID\t\tPPID\t\tTTY\tTIME\t\tCMD",
            "ppid": "Timestamp" + header_indentation + "PID\t\tPPID\t\tTTY\tTIME\t\tCMD",
            "username": "Timestamp" + header_indentation + "USERNAME\t\tPID\t\t%CPU\t%MEM\tVSZ\tRSS\t" +
                         "TTY\tSTAT\t\tTIME\t\tSTART\t\tCOMMAND",
            "command": "Timestamp" + header_indentation + "PID\t\tPPID\t\tTTY\tTIME\t\tCMD"
        }
        selected_flag = self.processStatOptions.universal_flag
        header = FORMAT_MAP.get(selected_flag, FORMAT_MAP["all"])
        # Bulk buffer for all rows, print all at once for efficiency
        output_rows = []
        cpu_idx , mem_idx = 0, 0
        processes = self.process_filter.filter_processes(self.process_report.get_processes(self.delta_time))
        def safe_str(val):
            return '' if val is None else str(val)
        if selected_flag == "all":
            for process in processes:
                output_rows.append("%s%s%s\t\t%s\t%s\t%s" % (
                    safe_str(timestamp), safe_str(value_indentation), safe_str(process.pid()),
                    safe_str(process.tty_name()), safe_str(process.total_time()),
                    safe_str(process.process_name_with_args(True))))
        elif selected_flag == "empty_arg":
            for process in processes:
                output_rows.append("%s%s%s\t%s\t%s" % (
                    safe_str(timestamp), safe_str(value_indentation), safe_str(process.pid()),
                    safe_str(process.total_time()), safe_str(process.process_name())))
        elif selected_flag == "pid":
            for process in processes:
                output_rows.append("%s%s%s\t%s\t%s\t%s\t%s" % (
                    safe_str(timestamp), safe_str(value_indentation), safe_str(process.pid()), safe_str(process.ppid()),
                    safe_str(process.tty_name()), safe_str(process.total_time()), safe_str(process.process_name())))
        elif selected_flag == "ppid":
            for process in processes:
                output_rows.append("%s%s%s\t%s\t%s\t%s\t%s" % (
                    safe_str(timestamp), safe_str(value_indentation), safe_str(process.pid()),
                    safe_str(process.ppid()), safe_str(process.tty_name()),
                    safe_str(process.total_time()), safe_str(process.process_name())))
        elif selected_flag == "username":
            for process in processes:
                output_rows.append("%s%s%s\t\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s" % (
                    safe_str(timestamp), safe_str(value_indentation), safe_str(process.user_name()),
                    safe_str(process.pid()),safe_str(process.system_percent()), safe_str(process.mem()),
                    safe_str(process.vsize()), safe_str(process.rss()),
                    safe_str(process.tty_name()), safe_str(process.ppid()), safe_str(process.total_time()),
                    safe_str(process.start()),
                    safe_str(process.process_name())))
            cpu_idx = 3
            mem_idx = 4
        elif selected_flag == "user":
            for process in processes:
                output_rows.append("%s%s%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s" % (
                    safe_str(timestamp), safe_str(value_indentation), safe_str(process.user_name()),
                    safe_str(process.pid()),
                    safe_str(process.system_percent()), safe_str(process.mem()),
                    safe_str(process.vsize()), safe_str(process.rss()),
                    safe_str(process.tty_name()), safe_str(process.s_name()),
                    safe_str(process.total_time()), safe_str(process.start()),
                    safe_str(process.process_name())))
            cpu_idx = 3
            mem_idx = 4
        elif selected_flag == "command":
            for process in processes:
                output_rows.append("%s%s%s\t%s\t%s\t%s\t%s" % (
                    safe_str(timestamp), safe_str(value_indentation), safe_str(process.pid()),
                    safe_str(process.ppid()), safe_str(process.tty_name()),
                    safe_str(process.total_time()), safe_str(process.process_name())))
        else:  # default fallback, print nothing extra
            pass
        if self.processStatOptions.sorting_flag:
            if cpu_idx == 0 and mem_idx == 0:
                raise ValueError("Sorting indices not set for selected flag, "
                                 "please remove sorting flag or choose another flag for output")
            if self.processStatOptions.sorting_order == '%cpu':
                output_rows.sort(
                key=lambda x: float(x.split()[cpu_idx]) if x.split()[cpu_idx].replace('.', '', 1).isdigit()
                else float('-inf'),
                reverse=True)
            elif self.processStatOptions.sorting_order == '%mem':
                output_rows.sort(
                key=lambda x: float(x.split()[mem_idx]) if x.split()[mem_idx].replace('.', '', 1).isdigit()
                else float('-inf'),
                reverse=True)
        if output_rows:
            self.printer(header)
            self.printer('\n'.join(output_rows))

class ProcessStatReport(pmcc.MetricGroupPrinter):
    Machine_info_count = 0
    group = None
    def __init__(self, group=None, options = None):
        self.group = group
        self.processStatOptions = options

    def timeStampDelta(self):
        s = self.group.timestamp.tv_sec - self.group.prevTimestamp.tv_sec
        n = self.group.timestamp.tv_nsec - self.group.prevTimestamp.tv_nsec
        return s + n / 1000000000.0

    def print_machine_info(self,context):
        timestamp = context.pmLocaltime(self.group.timestamp.tv_sec)
        # Please check strftime(3) for different formatting options.
        # Also check TZ and LC_TIME environment variables for more
        # information on how to override the default formatting of
        # the date display in the header
        time_string = time.strftime("%x", timestamp.struct_time())
        header_string = ''
        header_string += self.group['kernel.uname.sysname'].netValues[0][2] + '  '
        header_string += self.group['kernel.uname.release'].netValues[0][2] + '  '
        header_string += '(' + self.group['kernel.uname.nodename'].netValues[0][2] + ')  '
        header_string += time_string + '  '
        header_string += self.group['kernel.uname.machine'].netValues[0][2] + '  '
        print("%s  (%s CPU)" % (header_string, self.__get_ncpu(self.group)))

    def __get_ncpu(self, group):
        return group['hinv.ncpu'].netValues[0][2]

    def __print_report(self, manager,timestamp, header_indentation, value_indentation,interval_in_seconds):
        if self.processStatOptions.debug_mode:
            print("Printing standard report")
        metric_repository = ReportingMetricRepository(self.group)
        process_report = ProcessStatus(manager, metric_repository)
        process_filter = ProcessFilter(self.processStatOptions)
        stdout = StdoutPrinter()
        printdecorator = NoneHandlingPrinterDecorator(stdout)
        report = ProcessStatusReporter(process_report, process_filter, interval_in_seconds,
                                        printdecorator.Print, self.processStatOptions)
        report.print_report(timestamp, header_indentation, value_indentation)
    def __print_dynamic_report(self, manager,timestamp, header_indentation, value_indentation,interval_in_seconds):
        if self.processStatOptions.debug_mode:
            print("Printing dynamic report")
        metric_repository = ReportingMetricRepository(self.group)
        process_report = ProcessStatus(manager, metric_repository)
        process_filter = ProcessFilter(self.processStatOptions)
        stdout = StdoutPrinter()
        printdecorator = NoneHandlingPrinterDecorator(stdout)
        report = DynamicProcessReporter(process_report, process_filter, interval_in_seconds,
                                        printdecorator.Print, self.processStatOptions)
        report.print_report(timestamp, header_indentation, value_indentation)
    def __get_timestamp(self):
        ts = self.group.contextCache.pmLocaltime(int(self.group.timestamp))
        timestamp = time.strftime(self.processStatOptions.timefmt, ts.struct_time())
        return timestamp

    def report(self, manager):
        try:
            if self.group['proc.psinfo.utime'].netPrevValues is None:
                return False  # Not ready, skip increment
            if not self.group['hinv.ncpu'].netValues or not self.group['kernel.uname.sysname'].netValues:
                return False
            try:
                if not self.Machine_info_count:
                    self.print_machine_info(manager)
                    self.Machine_info_count = 1
            except IndexError:
                if self.processStatOptions.debug_mode:
                    print("IndexError while printing machine info")
                return False
            if self.processStatOptions.debug_mode:
                print("Starting report generation")
                print("Need to print samples: %s" % self.processStatOptions.print_count)
            if self.processStatOptions.print_count == 0:
                if self.processStatOptions.debug_mode:
                    print("Print count exhausted, exiting")
                sys.exit(0)
            timestamp = self.__get_timestamp()
            interval_in_seconds = self.timeStampDelta()
            header_indentation = "        " if len(timestamp) < 9 else (len(timestamp) - 7) * " "
            value_indentation = ((len(header_indentation) + 9) - len(timestamp)) * " "

            # Doing this for one single print instance in case there is no count specified
            if self.processStatOptions.print_count is None \
                and self.processStatOptions.context is not PM_CONTEXT_ARCHIVE:
                self.processStatOptions.print_count = 1
            # ================================================================
            if self.processStatOptions.selective_colum_flag:
                if self.processStatOptions.debug_mode:
                    print("Selective column flag is set")
                self.__print_dynamic_report(manager,timestamp, header_indentation,
                                            value_indentation, interval_in_seconds)
            else:
                self.__print_report(manager,timestamp, header_indentation, value_indentation, interval_in_seconds)
            if self.processStatOptions.context is not PM_CONTEXT_ARCHIVE:
                self.processStatOptions.print_count -= 1
            return True  # Data was printed
        finally:
            sys.stdout.flush()


class ProcessStatOptions(pmapi.pmOptions):
    universal_flag = None
    selective_colum_flag = False
    filter_flag = False
    filterstate = None
    debug_mode = False
    sorting_flag = False
    sorting_order = None
    timefmt = "%H:%M:%S"
    print_count = None
    column_list = []
    command_list = []
    pid_list = []
    ppid_list = []
    filtered_process_user = None
    context = None

    def __init__(self):
        pmapi.pmOptions.__init__(self, "t:c:e::p:ukVZ:z?:o:P:l:U:k:O:d")
        self.pmSetOptionCallback(self.extraOptions)
        self.pmSetOverrideCallback(self.override)
        self.options()

    def options(self):
        self.pmSetLongOptionHeader("General options")
        self.pmSetLongOption("", 0, "e", "",
                             "Show all process")
        self.pmSetLongOption("", 1, "c", "[Command name]",
                             "Show the stats for specified command name process")
        self.pmSetLongOption("", 1, "p", "[pid1,pid2,...]",
                             "Select the process by process ID")
        self.pmSetLongOption("", 1, "P", "[ppid1,ppid2,...]", "Select the process by process parent ID")
        self.pmSetLongOption("", 1, "U", "[User Name]", "Select the process by user name")
        self.pmSetLongOption(
            "",
            1,
            "o",
            "[col1,col2,... Or ALL]",
            (
                "User -defined format USE -o [all] or -B [col1, col2 , ...]\n"
                "\t\t\tsupported user defined colums are command, wchan, started, Time, pid, ppid, "
                "%mem, pri, user, %cpu and S\n"
                "\t\t\tALL option shows USER,PID,PPID,PRI,%CPU,%MEM,VSZ,RSS,S,STARTED,TIME,WCHAN and\n"
                "Command"
            ),
        )
        self.pmSetLongOptionText("\tCOL\tHEADER \tDESCRIPTION")
        self.pmSetLongOptionText("\t%cpu\t%CPU \tcpu utilization of the process")
        self.pmSetLongOptionText("\t%mem\t%MEM  \tphysical memory on the machine expressed as a percentage")
        self.pmSetLongOptionText("\tstart\tSTART \ttime the command started")
        self.pmSetLongOptionText("\ttime\tTIME \taccumulated cpu time, user + system")
        self.pmSetLongOptionText(
            "\tcls\tCLS \tscheduling class of the process"
        )
        self.pmSetLongOptionText(
            "\tcmd\tCMD\tsee args.  (alias args, command)."
        )
        self.pmSetLongOptionText(
            "\tpid\tPID \tthe process ID"
        )
        self.pmSetLongOptionText(
            "\tppid\tPPID\tparent process ID"
        )
        self.pmSetLongOptionText(
            "\tpri\tPRI \tpriority of the process"
        )
        self.pmSetLongOptionText(
            "\tstate\tS \tsee s"
        )
        self.pmSetLongOptionText(
            "\trss\tRSS \tthe non-swapped physical memory that a task has used"
        )
        self.pmSetLongOptionText(
            "\trtprio\tRTPRIO \trealtime priority"
        )
        self.pmSetLongOptionText(
            "\tpname\tPname\tProcess name"
        )
        self.pmSetLongOptionText(
            "\ttty\tTT \tcontrolling tty (terminal)"
        )
        self.pmSetLongOptionText(
            "\tuid\tUID \tsee euid"
        )
        self.pmSetLongOptionText(
            "\tvsize\tVSZ \tsee vsz"
        )
        self.pmSetLongOptionText(
            "\tuname\tUSER \tsee euser"
        )
        self.pmSetLongOptionText(
            "\twchan\tWCHAN \tname of the kernel function in which the process is sleeping"
        )
        self.pmSetLongOption("", 0, 'u', "",
                             "Display user-oriented format"
        )
        self.pmSetLongOption(
            "sort", 1, "O", "%cpu,%mem",
            "sort the process list by %cpu or %mem values "
        )
        self.pmSetLongOption("", 0, "d", "", "enable debug mode")
        self.pmSetLongOptionVersion()
        self.pmSetLongOptionTimeZone()
        self.pmSetLongOptionHostZone()
        self.pmSetLongOptionHelp()

    def override(self, opts):
        self.print_count = self.pmGetOptionSamples()
        # """Override standard Pcp-ps option to show all process """
        return bool(opts in ['p', 'c', 'o', 'P', 'U', 'O', 'd'])

    def extraOptions(self, opts, optarg, index):

        def handle_e():
            if self.debug_mode:
                print("e option selected")
            self.universal_flag = "all"

        def handle_c():
            if self.debug_mode:
                print("command option selected")
            self.universal_flag = "command"
            self.command_filter_flag = True
            self.filter_flag = True
            try:
                if optarg is not None:
                    self.command_list += optarg.replace(',', ' ').split(' ')
                    if self.debug_mode:
                        print("Command List: %s" % self.command_list)
            except ValueError:
                print("Invalid command Id List: use comma separated pids without whitespaces")
                sys.exit(1)
        def handle_d():
            print("Debug mode selected")
            self.debug_mode = True

        def handle_p():
            if self.debug_mode:
                print("pid option selected")
            self.universal_flag = "pid"
            self.filter_flag = True
            self.pid_filter_flag = True
            try:
                if optarg is not None:
                    dummy_list = optarg.replace(',', ' ').split(' ')
                    self.pid_list += [int(x) for x in dummy_list]
            except ValueError:
                print("Invalid pid Id List: use comma separated pids without whitespaces")
                sys.exit(1)

        def handle_P():
            if self.debug_mode:
                print("ppid option selected")
            self.universal_flag = "ppid"
            self.filter_flag = True
            self.ppid_filter_flag = True
            try:
                if optarg is not None:
                    dummy_list = optarg.replace(',', ' ').split(' ')
                    self.ppid_list += [int(x) for x in dummy_list]
            except ValueError:
                print("Invalid ppid Id List: use comma separated pids without whitespaces")
                sys.exit(1)

        def handle_u():
            if self.debug_mode:
                print("User-oriented format option selected")
            self.universal_flag = "user"
            self.user_oriented_format = True

        def handle_o():
            if self.debug_mode:
                print("User-defined format option selected")
            self.selective_colum_flag = True
            try:
                if optarg.upper() == "ALL":
                    self.filterstate = optarg.upper()
                    self.column_list = [
                        "uname", "pid", "ppid", "pri", "%cpu", "%mem",
                        "vsize", "rss", "state", "start", "time", "wchan", "args"
                    ]
                else:
                    dummy_list = optarg.replace(',', ' ').split(' ')
                    if self.debug_mode:
                        print("Custom Column List: %s" % dummy_list)
                    for key in dummy_list:
                        if key.lower() in PIDINFO_PAIR:
                            self.column_list.append(key.lower())
                        else:
                            raise ValueError
            except ValueError:
                print("Invalid Column List: incorrect name or improper comma-separated format")
                sys.exit(1)

        def handle_U():
            if self.debug_mode:
                print("username option selected")
            self.universal_flag = "username"
            self.username_filter_flag = True
            self.filter_flag = True
            self.filtered_process_user = optarg

        def handle_none():
            if self.debug_mode:
                print("No option selected, defaulting to show all process")
            self.universal_flag = "all"
            # self.show_all_process = True

        def handle_O():
            if self.debug_mode:
                print("Sorting option selected")
            self.sorting_flag = True
            if optarg.lower() not in ['%cpu', '%mem']:
                print("Invalid sorting option, defaulting to %cpu")
                self.sorting_order = '%cpu'  # default to %cpu
                return
            self.sorting_order = optarg.lower()

        # Dispatch map simulating switch-case
        dispatch = {
            'e': handle_e,
            'c': handle_c,
            'p': handle_p,
            'P': handle_P,
            'u': handle_u,
            'o': handle_o,
            'U': handle_U,
            'O': handle_O,
            'd': handle_d,
            None: handle_none,
        }
        # Execute handler or fallback to error
        handler = dispatch.get(opts)
        if handler:
            handler()
        else:
            print("Unknown option: %s" % opts)
            sys.exit(1)
    def checkOptions(self):
        if self.universal_flag is not None :
            return True
        else:
            if self.debug_mode:
                print("No filtering option selected, defaulting to empty argument mode")
            self.universal_flag = "empty_arg"
            return True


if __name__ == "__main__":
    try:
        opts = ProcessStatOptions()
        manager = pmcc.MetricGroupManager.builder(opts, sys.argv)
        ProcessStatOptions.context = manager.type
        if not opts.checkOptions():
            raise pmapi.pmUsageErr
        missing = manager.checkMissingMetrics(PSSTAT_METRICS)
        if missing is not None:
            sys.stderr.write('Error: not all required metrics are available\nMissing %s\n' % missing)
            sys.exit(1)
        manager['psstat'] = PSSTAT_METRICS
        manager.printer = ProcessStatReport(manager['psstat'],opts)
        sts = manager.run()
        sys.exit(sts)
    except pmapi.pmErr as pmerror:
        sys.stderr.write('%s\n' % (pmerror.message()))
    except pmapi.pmUsageErr as usage:
        usage.message()
        sys.exit(1)
    except ValueError as e:
        sys.stderr.write("%s\n" % str(e))
        sys.exit(1)
    except IOError:
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    except KeyboardInterrupt:
        print("Interrupted")
        sys.exit(0)



#!/usr/bin/env pmpython
#
# Copyright (c) 2022 Oracle and/or its affiliates.
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#

import mock
import unittest
from pcp_ps import ProcessStatusUtil


class TestProcessStackUtil(unittest.TestCase):
    def setUp(self):
        self.__metric_repository = mock.Mock()
        self.__metric_repository.current_value = mock.Mock(side_effect=self.metric_repo_current_value_side_effect)

    def metric_repo_current_value_side_effect(self, metric_name, instance):
        if metric_name == 'proc.psinfo.cmd' and instance == 1:
            return "test"
        if metric_name == 'proc.id.uid' and instance == 1:
            return 1
        if metric_name == 'proc.psinfo.pid' and instance == 1:
            return 1
        if metric_name == 'proc.id.uid_nm' and instance == 1:
            return "test"
        if metric_name == 'proc.psinfo.cmd' and instance == 1:
            return "test"
        if metric_name == 'proc.psinfo.psargs' and instance == 1:
            return "test"
        if metric_name == 'proc.psinfo.vsize' and instance == 1:
            return 1
        if metric_name == 'proc.psinfo.rss' and instance == 1:
            return 1
        if metric_name == 'mem.physmem' and instance is None:
            return 1
        if metric_name == 'proc.psinfo.sname' and instance == 1:
            return 'R'
        if metric_name == 'proc.psinfo.processor' and instance == 1:
            return 1
        if metric_name == 'proc.psinfo.stime' and instance == 1:
            return 1
        if metric_name == 'proc.psinfo.wchan_s' and instance == 1:
            return "test"
        if metric_name == 'proc.psinfo.priority' and instance == 1:
            return 1
        if metric_name == 'proc.psinfo.utime' and instance == 1:
            return 1
        if metric_name == 'proc.psinfo.guest_time' and instance == 1:
            return 1
        if metric_name == 'proc.psinfo.stime' and instance == 1:
            return 1
        if metric_name == 'proc.psinfo.start_time' and instance == 1:
            return 1
        if metric_name == 'kernel.all.uptime' and instance is None:
            return 1
        if metric_name == 'proc.psinfo.ttyname' and instance == 1:
            return 'tty'
        if metric_name == 'proc.psinfo.policy' and instance == 1:
            return 1

    def test_stack_referenced_size(self):
        self.skipTest(reason="Implement when suitable metric is found")

    #These are blank spaces in assert case been addded 
    #to match the format of function ouput.please don't remove
    def test_username(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.user_name()
        self.assertEqual(name, "test      ")

    def test_Processname(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.process_name()
        self.assertEqual(name, "test                ")

    def test_process_name_with_args(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.process_name_with_args()
        self.assertEqual(name, "test                          ")

    def test_vszie(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        vsize = process_status_usage.vsize()
        self.assertEqual(vsize, 1)

    def test_rss(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        vsize = process_status_usage.rss()
        self.assertEqual(vsize, 1)

    def test_mem(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        vsize = process_status_usage.mem()
        self.assertEqual(vsize, 100)

    def test_pid(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        pid = process_status_usage.pid()
        self.assertEqual(pid,'1       ')       

    def test_process_name(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.process_name()
        self.assertEqual(name, 'test                ')

    def test_user_id(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        user_id = process_status_usage.user_id()
        self.assertEqual(user_id, 1)

    def test_s_name(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.s_name()
        self.assertEqual(name, 'R')

    def test_cpu_number(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.cpu_number()
        self.assertEqual(name, 1)

    def test_wchan_s(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.wchan_s()
        self.assertEqual(name, 'test                          ')

    def test_priority(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.priority()
        self.assertEqual(name, 1)

    def test_tty_name(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.tty_name()
        self.assertEqual(name, 'tty')

    def test_start_time(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.start_time()
        self.assertEqual(name, 1)

    def test_func_state(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.func_state()
        self.assertEqual(name, 'N/A')

    def test_policy(self):
        process_status_usage = ProcessStatusUtil(1, 1.34, self.__metric_repository)
        name = process_status_usage.policy()
        self.assertEqual(name, 'FIFO')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
import argparse
import subprocess
import yaml
import sys
import os
import tempfile
import platform as platform_module
from datetime import datetime


def get_host_os():
    return platform_module.system()


def get_host_architecture():
    return platform_module.machine()


def is_macos():
    return get_host_os() == "Darwin"


def is_arm64():
    return get_host_architecture() in ("arm64", "aarch64")


def _parse_platform_string(value, separator):
    """Parse a string into a list of platform names, handling both delimiters and empty values."""
    if not value:
        return None
    # Use separator as primary delimiter, then split on whitespace
    platforms = [p.strip() for p in value.replace(separator, ' ').split() if p.strip()]
    return platforms if platforms else None


def resolve_platforms_to_run(cli_platforms=None, env_platforms=None, config_file=None):
    """
    Resolve platforms to run with priority hierarchy.

    Priority:
    1. CLI arguments (comma or space-separated)
    2. Environment variable PCP_CI_QUICK_PLATFORMS
    3. Config file .pcp-ci-quick (line-separated, comments ignored)
    4. None - requires user to specify platforms

    Args:
        cli_platforms: String of platforms from command line
        env_platforms: String of platforms from environment variable
        config_file: Path to config file (default: .pcp-ci-quick)

    Returns:
        List of platform names, or None if no platforms found
    """
    if config_file is None:
        config_file = ".pcp-ci-quick"

    # Priority 1: CLI arguments
    platforms = _parse_platform_string(cli_platforms, ',')
    if platforms:
        return platforms

    # Priority 2: Environment variable
    platforms = _parse_platform_string(env_platforms, ',')
    if platforms:
        return platforms

    # Priority 3: Config file
    if os.path.isfile(config_file):
        try:
            with open(config_file, encoding="utf-8") as f:
                platforms = [line.strip() for line in f if line.strip() and not line.strip().startswith('#')]
                if platforms:
                    return platforms
        except (IOError, OSError) as e:
            print(f"Warning: Could not read config file {config_file}: {e}", file=sys.stderr)

    return None


def print_quick_mode_help():
    """Print helpful error message for quick mode."""
    print("\nError: No platforms specified for --quick mode.", file=sys.stderr)
    print("\nPlease specify platforms using one of these methods:\n", file=sys.stderr)
    print("  1. Command line (comma or space-separated):", file=sys.stderr)
    print("     python3 build/ci/ci-run.py --quick ubuntu2404-container,fedora43-container reproduce\n", file=sys.stderr)
    print("  2. Environment variable:", file=sys.stderr)
    print("     export PCP_CI_QUICK_PLATFORMS='ubuntu2404-container fedora43-container'", file=sys.stderr)
    print("     python3 build/ci/ci-run.py --quick reproduce\n", file=sys.stderr)
    print("  3. Config file (.pcp-ci-quick in repo root):", file=sys.stderr)
    print("     ubuntu2404-container", file=sys.stderr)
    print("     fedora43-container", file=sys.stderr)
    print("     centos-stream10-container", file=sys.stderr)
    print("     python3 build/ci/ci-run.py --quick reproduce\n", file=sys.stderr)


def _execute_command(runner, args, platform_name=None):
    """
    Execute the appropriate command on the runner.

    Args:
        runner: The runner instance (ContainerRunner, etc.)
        args: Parsed command-line arguments
        platform_name: Optional platform name for logging (used in quick mode)
    """
    try:
        if args.main_command == "setup":
            runner.setup(args.pcp_path)
            runner.task("setup")
        elif args.main_command == "destroy":
            runner.destroy()
        elif args.main_command == "task":
            runner.task(args.task_name)
        elif args.main_command == "artifacts":
            runner.task(f"copy_{args.artifact}_artifacts")
            runner.get_artifacts(args.artifact, args.path)
        elif args.main_command == "exec":
            runner.exec(" ".join(args.command), check=False)
        elif args.main_command == "shell":
            runner.shell()
        elif args.main_command == "reproduce":
            all_tasks = list(runner.platform["tasks"].keys())
            if args.until not in all_tasks:
                print(f"Error: Unknown task '{args.until}'. Available tasks: {', '.join(all_tasks)}", file=sys.stderr)
                sys.exit(1)
            run_tasks = all_tasks[: all_tasks.index(args.until) + 1]

            if platform_name:
                # In quick mode, shorten the message
                print(f"[{platform_name}] Running tasks: {', '.join(run_tasks)}")
            else:
                print("Preparing a new virtual environment with PCP preinstalled, this will take about 20 minutes...")

            started = datetime.now()
            runner.setup(args.pcp_path)
            for task in run_tasks:
                print(f"\n[{platform_name if platform_name else 'CI'}] Running task {task}...")
                runner.task(task)
            duration_min = (datetime.now() - started).total_seconds() / 60
            print(f"\n[{platform_name if platform_name else 'CI'}] Tasks completed, took {duration_min:.0f}m.")

            if not platform_name:  # Only show in non-quick mode
                if "install" in all_tasks and all_tasks.index(args.until) >= all_tasks.index("install"):
                    print("\nPlease run:\n")
                    print("    sudo -u pcpqa -i ./check XXX\n")
                    print("to run a QA test. PCP is already installed, from sources located in './pcp'.")
                print("Starting a shell in the new virtual environment...\n")
                runner.shell()
        else:
            print(f"Error: Unknown command {args.main_command}", file=sys.stderr)
            sys.exit(1)
    except subprocess.CalledProcessError as e:
        print(f"Error on {platform_name or 'command'}: {e}", file=sys.stderr)
        # In quick mode, continue to next platform instead of exiting
        if platform_name:
            return
        sys.exit(1)


class DirectRunner:
    def __init__(self, platform_name: str, platform):
        self.platform_name = platform_name
        self.platform = platform
        # example: /tmp/pcp-ci-ubuntu2004
        self.build_dir = os.path.join(tempfile.gettempdir(), f"pcp-ci-{self.platform_name}")
        self.command_preamble = "set -eux\nexport runner=direct\n"

    def setup(self, pcp_path):
        os.mkdir(self.build_dir)
        # copy sources to a temp folder to not mess up the source dir
        subprocess.run(["rsync", "-a", f"{pcp_path}/", f"{self.build_dir}/pcp/"], check=True)
        self.exec("mkdir -p ../artifacts/build ../artifacts/test")

    def destroy(self):
        pass

    def exec(self, command, check=True):
        command = self.command_preamble + command
        subprocess.run(["bash", "-"], cwd=f"{self.build_dir}/pcp", input=command.encode(), check=check)

    def shell(self):
        pass

    def task(self, task_name):
        self.exec(self.platform["tasks"][task_name])

    def get_artifacts(self, artifact, path):
        subprocess.run(
            [
                "rsync",
                "-a",
                f"{self.build_dir}/artifacts/{artifact}/",
                f"{path}/",
            ],
            check=True,
        )


class VirtualMachineRunner:
    """
    this class is deprecated and might get removed soon
    """

    def __init__(self, platform_name: str, platform):
        self.platform_name = platform_name
        self.platform = platform
        self.vm_name = f"pcp-ci-{self.platform_name}"
        self.ssh_config_file = f".ssh-config-{self.platform_name}"
        self.vagrant_env = {
            "PATH": "/usr/bin",
            "VAGRANT_CWD": os.path.dirname(__file__),
            "VAGRANT_NAME": self.vm_name,
            "VAGRANT_BOX": self.platform["vm"]["box"],
        }
        self.command_preamble = "set -eux\nexport runner=vm\n"

    def setup(self, pcp_path):
        post_start = self.platform["container"].get("post_start")

        subprocess.run(["vagrant", "up"], env=self.vagrant_env, check=True)
        subprocess.run(f"vagrant ssh-config > {self.ssh_config_file}", env=self.vagrant_env, shell=True, check=True)
        subprocess.run(
            ["rsync", "-a", "-e", f"ssh -F {self.ssh_config_file}", f"{pcp_path}/", f"{self.vm_name}:pcp/"], check=True
        )
        self.exec("mkdir -p artifacts/build artifacts/test")

        if post_start:
            self.exec(post_start)

    def exec(self, command, check=True):
        command = self.command_preamble + command
        subprocess.run(
            ["ssh", "-F", self.ssh_config_file, self.vm_name, "bash", "-"], input=command.encode(), check=check
        )

    def shell(self):
        subprocess.run(["ssh", "-F", self.ssh_config_file, self.vm_name], check=False)

    def task(self, task_name):
        self.exec(self.platform["tasks"][task_name])

    def get_artifacts(self, artifact, path):
        subprocess.run(
            [
                "rsync",
                "-a",
                "-e",
                f"ssh -F {self.ssh_config_file}",
                f"{self.vm_name}:artifacts/{artifact}/",
                f"{path}/",
            ],
            check=True,
        )


class ContainerRunner:
    def _setup_macos_config(self):
        """Configure podman for macOS (no sudo, platform flags for ARM64)."""
        self.sudo = []
        self.security_opts = []
        if is_arm64():
            # Specify container architecture explicitly on ARM64 macOS
            arch = "linux/arm64" if self.use_native_arch else "linux/amd64"
            self.platform_flags = ["--platform", arch]

    def _setup_linux_config(self):
        """Configure podman for Linux (sudo for Ubuntu, system labels)."""
        try:
            with open("/etc/os-release", encoding="utf-8") as f:
                for line in f:
                    k, v = line.rstrip().split("=")
                    if k == "NAME" and v == '"Ubuntu"':
                        self.sudo = ["sudo", "-E", "XDG_RUNTIME_DIR="]
                        self.security_opts = ["--security-opt", "label=disable"]
                        break
        except FileNotFoundError:
            pass

    def __init__(self, platform_name: str, platform, use_native_arch: bool = False):
        self.platform_name = platform_name
        self.platform = platform
        self.container_name = f"pcp-ci-{self.platform_name}"
        self.image_name = f"{self.container_name}-image"
        self.command_preamble = "set -eux\nexport runner=container\n"
        self.platform_flags = []
        self.use_native_arch = use_native_arch
        self.sudo = []
        self.security_opts = []

        if is_macos():
            self._setup_macos_config()
        else:
            self._setup_linux_config()

    def setup(self, pcp_path):
        containerfile = self.platform["container"]["containerfile"]

        # platform_flags specifies container architecture (e.g., --platform linux/arm64)
        # on ARM64 macOS, allowing explicit control over native vs emulated builds
        subprocess.run(
            [*self.sudo, "podman", "build", *self.platform_flags, "--squash", "-t", self.image_name, "-f", "-"],
            input=containerfile.encode(),
            check=True,
        )

        subprocess.run([*self.sudo, "podman", "rm", "-f", self.container_name], stderr=subprocess.DEVNULL, check=False)
        subprocess.run(
            [
                *self.sudo,
                "podman",
                "run",
                *self.platform_flags,
                "-dt",
                "--name",
                self.container_name,
                "--privileged",
                *self.security_opts,
                self.image_name,
            ],
            check=True,
        )

        # Copy PCP sources including .git directory
        # Makepkgs requires a valid git repository for status, checkout, and archive operations
        subprocess.run(
            [*self.sudo, "podman", "cp", f"{pcp_path}/.", f"{self.container_name}:/home/pcpbuild/pcp"], check=True
        )

        self.exec("sudo chown -R pcpbuild:pcpbuild .")
        self.exec("mkdir -p ../artifacts/build ../artifacts/test")

    def destroy(self):
        subprocess.run([*self.sudo, "podman", "rm", "-f", self.container_name], check=True)
        subprocess.run([*self.sudo, "podman", "rmi", self.image_name], check=True)

    def exec(self, command, check=True):
        command = self.command_preamble + command
        subprocess.run(
            [
                *self.sudo,
                "podman",
                "exec",
                "-i",
                "-u",
                "pcpbuild",
                "-w",
                "/home/pcpbuild/pcp",
                "-e",
                "PCP_QA_ARGS",
                self.container_name,
                "bash",
                "-",
            ],
            input=command.encode(),
            check=check,
        )

    def shell(self):
        subprocess.run(
            [
                *self.sudo,
                "podman",
                "exec",
                "-it",
                "-u",
                "pcpbuild",
                "-w",
                "/home/pcpbuild/pcp",
                self.container_name,
                "bash",
            ],
            check=False,
        )

    def task(self, task_name):
        self.exec(self.platform["tasks"][task_name])

    def get_artifacts(self, artifact, path):
        subprocess.run(
            [*self.sudo, "podman", "cp", f"{self.container_name}:/home/pcpbuild/artifacts/{artifact}/.", path],
            check=True,
        )


def _determine_use_native_arch(args):
    """Determine whether to use native architecture based on platform and flags."""
    use_native_arch = False
    if is_macos():
        # On macOS: default to native arch unless --emulate is specified
        use_native_arch = not args.emulate
    # On Linux or with --native-arch flag: respect explicit --native-arch
    if args.native_arch:
        use_native_arch = True
    return use_native_arch


def _create_runner(platform_name, platform, use_native_arch):
    """Create the appropriate runner for the given platform type."""
    platform_type = platform.get("type")
    if platform_type == "direct":
        return DirectRunner(platform_name, platform)
    elif platform_type == "vm":
        return VirtualMachineRunner(platform_name, platform)
    elif platform_type == "container":
        return ContainerRunner(platform_name, platform, use_native_arch=use_native_arch)
    else:
        print(f"Error: Unknown platform type: {platform_type}", file=sys.stderr)
        sys.exit(1)


def _load_platform_definition(platform_name):
    """Load platform YAML definition file."""
    platform_def_path = os.path.join(os.path.dirname(__file__), f"platforms/{platform_name}.yml")
    try:
        with open(platform_def_path, encoding="utf-8") as f:
            return yaml.safe_load(f)
    except FileNotFoundError:
        print(f"Error: Platform definition not found: {platform_def_path}", file=sys.stderr)
        sys.exit(1)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--pcp_path", default=".")
    parser.add_argument(
        "--native-arch",
        action="store_true",
        help="Use native host architecture instead of amd64 (useful on macOS for faster builds)"
    )
    parser.add_argument(
        "--emulate",
        action="store_true",
        help="Force amd64 emulation even on native ARM64 systems (default on non-macOS)"
    )
    parser.add_argument(
        "--quick",
        nargs="?",
        const=True,
        metavar="PLATFORMS",
        help="Quick mode: run multiple platforms. Platforms can be comma or space-separated, "
             "or loaded from PCP_CI_QUICK_PLATFORMS env var or .pcp-ci-quick config file"
    )
    parser.add_argument("platform", nargs="?")
    subparsers = parser.add_subparsers(dest="main_command")

    subparsers.add_parser("setup")
    subparsers.add_parser("destroy")

    parser_task = subparsers.add_parser("task")
    parser_task.add_argument("task_name")

    parser_artifacts = subparsers.add_parser("artifacts")
    parser_artifacts.add_argument("artifact", choices=["build", "test"])
    parser_artifacts.add_argument("--path", default="./artifacts")

    parser_exec = subparsers.add_parser("exec")
    parser_exec.add_argument("command", nargs=argparse.REMAINDER)

    subparsers.add_parser("shell")

    parser_reproduce = subparsers.add_parser("reproduce")
    parser_reproduce.add_argument("--until", default="init_qa")

    args = parser.parse_args()

    # Handle quick mode
    if args.quick is not None:
        quick_platforms = resolve_platforms_to_run(
            cli_platforms=args.platform if args.quick is True else args.quick,
            env_platforms=os.environ.get("PCP_CI_QUICK_PLATFORMS")
        )

        if not quick_platforms:
            print_quick_mode_help()
            sys.exit(1)

        if not args.main_command:
            print("Error: Quick mode requires a subcommand (setup, task, reproduce, etc.)", file=sys.stderr)
            sys.exit(1)

        use_native_arch = _determine_use_native_arch(args)

        for platform_name in quick_platforms:
            print(f"\n{'='*60}")
            print(f"Running on platform: {platform_name}")
            print(f"{'='*60}\n")

            platform = _load_platform_definition(platform_name)
            runner = _create_runner(platform_name, platform, use_native_arch)
            _execute_command(runner, args, platform_name)

        sys.exit(0)

    # Normal (non-quick) mode
    if not args.platform:
        print("Error: Platform argument required (unless using --quick mode)", file=sys.stderr)
        sys.exit(1)

    platform = _load_platform_definition(args.platform)
    use_native_arch = _determine_use_native_arch(args)
    runner = _create_runner(args.platform, platform, use_native_arch)

    _execute_command(runner, args)


if __name__ == "__main__":
    main()



#!/usr/bin/pmpython
#
# Copyright (c) 2023 Oracle and/or its affiliates.
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# pylint: disable=bad-whitespace,too-many-lines,bad-continuation
# pylint: disable=too-many-arguments,too-many-positional-arguments
# pylint: disable=redefined-outer-name,unnecessary-lambda
#
import signal
import sys
import time
from typing import Dict
from typing import Union

from cpmapi import PM_CONTEXT_ARCHIVE
from pcp import pmapi
from pcp import pmcc

SYS_METRICS = [
    "kernel.uname.sysname",
    "kernel.uname.release",
    "kernel.uname.nodename",
    "kernel.uname.machine",
    "hinv.ncpu",
]
NFSIOSTAT_METRICS = [
    "nfsclient.mountpoint",
    "nfsclient.export",
    "nfsclient.age",
    "nfsclient.xprt.sends",
    "nfsclient.xprt.backlog_u",
    "nfsclient.ops.read.ops",
    "nfsclient.ops.read.errors",
    "nfsclient.ops.read.execute",
    "nfsclient.ops.read.rtt",
    "nfsclient.ops.read.queue",
    "nfsclient.ops.read.bytes_recv",
    "nfsclient.ops.read.bytes_sent",
    "nfsclient.ops.read.ntrans",
    "nfsclient.ops.write.ops",
    "nfsclient.ops.write.errors",
    "nfsclient.ops.write.execute",
    "nfsclient.ops.write.rtt",
    "nfsclient.ops.write.queue",
    "nfsclient.ops.write.bytes_recv",
    "nfsclient.ops.write.bytes_sent",
    "nfsclient.ops.write.ntrans",
]
ALL_METRICS = NFSIOSTAT_METRICS + SYS_METRICS


def adjust_length(name):
    return name.ljust(25)


class ReportingMetricRepository:
    def __init__(self, group):
        self.group = group
        self._current_cache = {}
        self._previous_cache = {}

    def _fetch_values(self, metric, use_previous=False):
        """Fetch values - always returns a dictionary."""
        if metric not in self.group:
            return {}
        attr = "netPrevValues" if use_previous else "netValues"
        values = getattr(self.group[metric], attr, [])
        return {x[0].inst: x[2] for x in values} if values else {}

    def _get_values_dict(self, metric, use_previous=False):
        """Get cached dictionary of all values for a metric."""
        cache = self._previous_cache if use_previous else self._current_cache
        if metric not in cache:
            cache[metric] = self._fetch_values(metric, use_previous)
        return cache[metric]

    def previous_value(self, metric, instance=None):
        """Get previous value. Returns single value if instance given, else returns dict."""
        values_dict = self._get_values_dict(metric, use_previous=True)
        if instance is not None:
            return values_dict.get(instance)
        return values_dict

    def current_value(self, metric, instance=None):
        """Get current value. Returns single value if instance given, else returns dict."""
        values_dict = self._get_values_dict(metric, use_previous=False)
        if instance is not None:
            return values_dict.get(instance)
        return values_dict


class NfsioStatUtil:
    def __init__(self, metrics_repository):
        self.__metric_repository = metrics_repository
        self.report = ReportingMetricRepository(self.__metric_repository)


class NfsiostatReport(pmcc.MetricGroupPrinter):
    machine_info_count = 0

    def __init__(self, opts, group):
        self.opts = opts
        self.group = group
        self.samples = opts.samples
        self.context = opts.context

    def __get_ncpu(self, group):
        return group["hinv.ncpu"].netValues[0][2]

    def __print_machine_info(self, context):
        timestamp = self.group.pmLocaltime(context.timestamp.tv_sec)
        # Please check strftime(3) for different formatting options.
        # Also check TZ and LC_TIME environment variables for more
        # information on how to override the default formatting of
        # the date display in the header
        time_string = time.strftime(
            "%m/%d/%Y %H:%M:%S", timestamp.struct_time()
        )
        header_string = ""
        header_string += context["kernel.uname.sysname"].netValues[0][2] + "  "
        header_string += context["kernel.uname.release"].netValues[0][2] + "  "
        header_string += (
            "(" + context["kernel.uname.nodename"].netValues[0][2] + ")  "
        )
        header_string += time_string + "  "
        header_string += context["kernel.uname.machine"].netValues[0][2] + "  "
        print("%s  (%s CPU)" % (header_string, self.__get_ncpu(context)))

    def __collect(self, nfs):
        return {
            metric: nfs.report.current_value(metric)
            for metric in NFSIOSTAT_METRICS
        }

    def __delta(
        self,
        new: Dict[str, Dict[str, Union[float, str]]],
        nfs,
    ) -> Dict[str, Dict[str, Union[float, str]]]:
        delta: Dict[str, Dict[str, Union[float, str]]] = {}
        old = {
            metric: nfs.report.previous_value(metric)
            for metric in NFSIOSTAT_METRICS
        }

        for metric in new:
            delta[metric] = {}

            for inst in new[metric]:
                new_val = new[metric][inst]
                old_val = old.get(metric, {}).get(inst, 0)

                # If value is numeric → subtract
                if isinstance(new_val, (int, float)):
                    delta[metric][inst] = new_val - old_val
                else:
                    # If string → just copy (no subtraction)
                    delta[metric][inst] = new_val

        return delta

    def __print_values(self, timestamp, delta):
        sampletime = delta["nfsclient.age"]
        readops = delta["nfsclient.ops.read.ops"]
        writeops = delta["nfsclient.ops.write.ops"]
        readbytesrecv = delta["nfsclient.ops.read.bytes_recv"]
        writebytesrecv = delta["nfsclient.ops.write.bytes_recv"]
        mountpoint = delta["nfsclient.mountpoint"]
        mountshare = delta["nfsclient.export"]
        sends = delta["nfsclient.xprt.sends"]
        backlog = delta["nfsclient.xprt.backlog_u"]
        readerrors = delta["nfsclient.ops.read.errors"]
        readexecute = delta["nfsclient.ops.read.execute"]
        readrtt = delta["nfsclient.ops.read.rtt"]
        readqueue = delta["nfsclient.ops.read.queue"]
        readbytessent = delta["nfsclient.ops.read.bytes_sent"]
        readntrans = delta["nfsclient.ops.read.ntrans"]
        writeerrors = delta["nfsclient.ops.write.errors"]
        writeexecute = delta["nfsclient.ops.write.execute"]
        writertt = delta["nfsclient.ops.write.rtt"]
        writequeue = delta["nfsclient.ops.write.queue"]
        writebytessent = delta["nfsclient.ops.write.bytes_sent"]
        writentrans = delta["nfsclient.ops.write.ntrans"]

        def _fit(text, width):
            text = str(text)
            if len(text) <= width:
                return text.rjust(width)
            return "#" * width

        def _fmt_float(value, width, precision=3):
            if not isinstance(value, (int, float)):
                return _fit(value, width)
            fixed = f"{value:.{precision}f}"
            if len(fixed) <= width:
                return fixed.rjust(width)
            exp_precision = max(0, width - 7)
            scientific = f"{value:.{exp_precision}e}"
            if len(scientific) <= width:
                return scientific.rjust(width)
            return _fit(scientific, width)

        def _fmt_count_percent(count, percent, width):
            try:
                count_value = int(count)
            except (TypeError, ValueError):
                count_value = 0
            try:
                percent_value = float(percent)
            except (TypeError, ValueError):
                percent_value = 0.0

            formatted = f"{count_value} ({percent_value:.1f}%)"
            if len(formatted) <= width:
                return formatted.rjust(width)

            formatted = f"{count_value} ({percent_value:.1e}%)"
            if len(formatted) <= width:
                return formatted.rjust(width)

            exp_precision = max(0, width - 7)
            formatted = f"{float(count_value):.{exp_precision}e}"
            if len(formatted) <= width:
                return formatted.rjust(width)

            return _fit(formatted, width)

        io_columns = [
            ("ops/s", 11),
            ("kB/s", 12),
            ("kB/op", 13),
            ("retrans", 16),
            ("avg RTT (ms)", 15),
            ("avg exe (ms)", 15),
            ("avg queue (ms)", 17),
            ("errors", 14),
        ]
        rpc_columns = [("ops/s", 11), ("rpc bklog", 16)]

        print("%-18s:%s" % ("Timestamp", timestamp))
        print()

        for name in mountshare:
            # read
            r_kilobytes = (readbytessent[name] + readbytesrecv[name]) / 1024
            if sampletime[name] > 0:
                ops_per_sample = sends[name] / sampletime[name]
                ops_per_sample_read = readops[name] / sampletime[name]
                r_kilobytes_per_sample = r_kilobytes / sampletime[name]
            else:
                ops_per_sample = 0.0
                ops_per_sample_read = 0.0
                r_kilobytes_per_sample = 0.0

            r_retrans = readntrans[name] - readops[name]
            if readops[name] > 0:
                r_kilobytes_per_op = r_kilobytes / readops[name]
                r_retrans_percent = (r_retrans * 100) / readops[name]
                r_rtt_per_op = readrtt[name] / readops[name]
                r_exe_per_op = readexecute[name] / readops[name]
                r_queued_for_per_op = readqueue[name] / readops[name]
                r_errs_percent = (readerrors[name] * 100) / readops[name]
            else:
                r_kilobytes_per_op = 0.0
                r_retrans_percent = 0.0
                r_rtt_per_op = 0.0
                r_exe_per_op = 0.0
                r_queued_for_per_op = 0.0
                r_errs_percent = 0.0

            # write
            w_kilobytes = (writebytessent[name] + writebytesrecv[name]) / 1024
            if sampletime[name] > 0:
                ops_per_sample_write = writeops[name] / sampletime[name]
                w_kilobytes_per_sample = w_kilobytes / sampletime[name]
            else:
                ops_per_sample_write = 0.0
                w_kilobytes_per_sample = 0.0

            w_retrans = writentrans[name] - writeops[name]
            if writeops[name] > 0:
                w_kilobytes_per_op = w_kilobytes / writeops[name]
                w_retrans_percent = (w_retrans * 100) / writeops[name]
                w_rtt_per_op = writertt[name] / writeops[name]
                w_exe_per_op = writeexecute[name] / writeops[name]
                w_queued_for_per_op = writequeue[name] / writeops[name]
                w_errs_percent = (writeerrors[name] * 100) / writeops[name]
            else:
                w_kilobytes_per_op = 0.0
                w_retrans_percent = 0.0
                w_rtt_per_op = 0.0
                w_exe_per_op = 0.0
                w_queued_for_per_op = 0.0
                w_errs_percent = 0.0

            print(f"{mountshare[name]} mounted on {mountpoint[name]}:")

            print(f"{'':14}" + "".join(f"{title:>{width}}" for title, width in rpc_columns))
            print(f"{'':14}{_fmt_float(ops_per_sample, 11)}{_fmt_float(backlog[name], 16)}")
            print()
            print(
                "read:              "
                + "".join(f"{title:>{width}}" for title, width in io_columns)
            )
            print(
                f"{'':19}"
                f"{_fmt_float(ops_per_sample_read, 11)}"
                f"{_fmt_float(r_kilobytes_per_sample, 12)}"
                f"{_fmt_float(r_kilobytes_per_op, 13)}"
                f"{_fmt_count_percent(r_retrans, r_retrans_percent, 16)}"
                f"{_fmt_float(r_rtt_per_op, 15)}"
                f"{_fmt_float(r_exe_per_op, 15)}"
                f"{_fmt_float(r_queued_for_per_op, 17)}"
                f"{_fmt_count_percent(readerrors[name], r_errs_percent, 14)}"
            )

            print(
                "write:             "
                + "".join(f"{title:>{width}}" for title, width in io_columns)
            )
            print(
                f"{'':19}"
                f"{_fmt_float(ops_per_sample_write, 11)}"
                f"{_fmt_float(w_kilobytes_per_sample, 12)}"
                f"{_fmt_float(w_kilobytes_per_op, 13)}"
                f"{_fmt_count_percent(w_retrans, w_retrans_percent, 16)}"
                f"{_fmt_float(w_rtt_per_op, 15)}"
                f"{_fmt_float(w_exe_per_op, 15)}"
                f"{_fmt_float(w_queued_for_per_op, 17)}"
                f"{_fmt_count_percent(writeerrors[name], w_errs_percent, 14)}"
            )
            print()

    def get_timestamp(self, group):
        t_s = group.contextCache.pmLocaltime(int(group.timestamp))
        timestamp = time.strftime(NfsiostatOptions.timefmt, t_s.struct_time())
        return timestamp

    def print_report(self, manager_nfsiostat, mgr):
        def __print_nfs_status():
            timestamp = self.get_timestamp(manager_nfsiostat)
            nfs = NfsioStatUtil(manager_nfsiostat)
            if nfs.report.current_value("nfsclient.export"):
                try:
                    if self.machine_info_count == 0:
                        self.__print_machine_info(manager_nfsiostat)
                        self.machine_info_count = 1
                    current = self.__collect(nfs)
                    delta = self.__delta(current, nfs)
                    self.__print_values(timestamp, delta)

                except IndexError:
                    print("Incorrect machine info due to some missing metrics")
                return
            else:
                pass

        if self.context != PM_CONTEXT_ARCHIVE and self.samples is None:
            __print_nfs_status()
        elif self.context == PM_CONTEXT_ARCHIVE and self.samples is None:
            __print_nfs_status()
        elif self.samples >= 1:
            __print_nfs_status()
            self.samples -= 1
        else:
            pass

    def report(self, manager):
        self.samples = self.opts.pmGetOptionSamples()
        self.print_report(manager["nfsiostat"], manager)


class NfsiostatOptions(pmapi.pmOptions):
    timefmt = "%m/%d/%Y %H:%M:%S"

    def __init__(self):
        pmapi.pmOptions.__init__(self, "a:h:O:S:s:T:t:VZ:z?")
        self.pmSetLongOptionHeader("General options")
        self.pmSetLongOptionHost()
        self.pmSetLongOptionHostZone()
        self.pmSetLongOptionTimeZone()
        self.pmSetLongOptionArchive()
        self.pmSetLongOptionOrigin()
        self.pmSetLongOptionStart()
        self.pmSetLongOptionSamples()
        self.pmSetLongOptionFinish()
        self.pmSetLongOptionInterval()
        self.pmSetLongOptionHelp()
        self.pmSetLongOptionVersion()
        self.context = None
        self.samples = None


if __name__ == "__main__":
    try:
        opts = NfsiostatOptions()
        mngr = pmcc.MetricGroupManager.builder(opts, sys.argv)
        opts.context = mngr.type

        missing = mngr.checkMissingMetrics(ALL_METRICS)
        if missing is not None:
            sys.stderr.write(
                "Error: not all required metrics are available\nMissing %s\n"
                % missing
            )
            sys.exit(1)
        mngr["nfsiostat"] = ALL_METRICS
        mngr.printer = NfsiostatReport(opts, mngr)
        sts = mngr.run()
        sys.exit(sts)
    except pmapi.pmErr as error:
        sys.stderr.write("%s\n" % (error.message()))
    except pmapi.pmUsageErr as usage:
        usage.message()
        sys.exit(1)
    except IOError:
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    except KeyboardInterrupt:
        pass
